# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the database_interface module within the ketos library
"""
import pytest
import tables
import os
import ketos.data_handling.database_interface as di
import ketos.data_handling.data_handling as dh
from ketos.audio_processing.spectrogram import MagSpectrogram, Spectrogram, CQTSpectrogram


current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


@pytest.mark.test_open_table
def test_open_non_existing_table():
    """ Test if the expected exception is raised when the table does not exist """
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp1_db.h5')
    h5file = tables.open_file(fpath, 'w')
    # open non-existing table
    with pytest.raises(tables.NoSuchNodeError):
        tbl = di.open_table(h5file=h5file, table_path='/group_1/table_1')
        assert tbl == None
    # clean
    h5file.close()
    os.remove(fpath)

@pytest.mark.test_open_table
def test_open_existing_table():
    """ Test if the expected table is open """
    # open h5 file
    fpath = os.path.join(path_to_assets, '15x_same_spec.h5')
    h5file = tables.open_file(fpath, 'r')
    # open non-existing table
    tbl = di.open_table(h5file=h5file, table_path='/train/species1')
    assert isinstance(tbl, tables.table.Table)
    assert tbl.nrows == 15
    # clean
    h5file.close()
   

@pytest.mark.test_create_table
def test_create_table():
    """Test if a table and its group are created"""
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp2_db.h5')
    h5file = tables.open_file(fpath, 'w')
    # create table
    _ = di.create_table(h5file=h5file, path='/group_1/', name='table_1', shape=(20,60))
    group = h5file.get_node("/group_1")
    assert isinstance(group, tables.group.Group)
    table = h5file.get_node("/group_1/table_1")
    assert isinstance(table, tables.table.Table)    
    # clean
    h5file.close()
    os.remove(fpath)

@pytest.mark.test_create_table
def test_create_table_existing():
    """Test if a table is open when it already exists"""
    # open h5 file
    fpath = os.path.join(path_to_assets, '15x_same_spec.h5')
    h5file = tables.open_file(fpath, 'a')
    # create table
    _ = di.create_table(h5file=h5file, path='/train/', name='species1', shape=(20,60))
    table = h5file.get_node("/train/species1")
    assert table.nrows == 15
    assert table[0]['data'].shape == (2413,201)
    assert table[1]['id'] == b'1'
    # clean
    h5file.close()
    

@pytest.mark.test_write_spec
def test_write_spec(sine_audio):
    """Test if spectrograms are written and have the expected ids"""
    # create spectrogram    
    spec = MagSpectrogram(sine_audio, 0.5, 0.1)
    # add annotation
    spec.annotate(labels=(1,2), boxes=((1,2,3,4),(1.5,2.5,3.5,4.5)))
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp3_db.h5')
    h5file = tables.open_file(fpath, 'w')
    # create table
    tbl = di.create_table(h5file=h5file, path='/group_1/', name='table_1', shape=spec.image.shape)
    # write spectrogram to table
    di.write_spec(table=tbl, spec=spec) # should return None after writing spectrogram to table
    # write spectrogram to table with id
    di.write_spec(table=tbl, spec=spec, id='123%')

    assert tbl[0]['id'].decode() == ''
    assert tbl[0]['labels'].decode() == '[1,2]'
    assert tbl[0]['boxes'].decode() == '[[1.0,2.0,3.0,4.0],[1.5,2.5,3.5,4.5]]'

    assert tbl[1]['id'].decode() == '123%'
    assert tbl[1]['labels'].decode() == '[1,2]'
    assert tbl[1]['boxes'].decode() == '[[1.0,2.0,3.0,4.0],[1.5,2.5,3.5,4.5]]'

    h5file.close()
    os.remove(fpath)

@pytest.mark.test_write_spec
def test_write_spec_TypeError(sine_audio):
    """Test if a type error is raised when trying to pass an object that isn't an instance of Spectrogram (or its subclasses)"""
    sine_audio.annotate(labels=(1,2), boxes=((1,2,3,4),(1.5,2.5,3.5,4.5)))
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp4_db.h5')
    h5file = tables.open_file(fpath, 'w')
    # create table
    tbl = di.create_table(h5file=h5file, path='/group_1/', name='table_1', shape=sine_audio.data.shape)
    # write spectrogram to table
    with pytest.raises(TypeError):
        di.write_spec(table=tbl, spec=sine_audio)
       
    assert tbl.nrows == 0
    
    h5file.close()
    os.remove(fpath)

@pytest.mark.test_write_spec
def test_write_spec_cqt(sine_audio):
    """Test if CQT spectrograms are written with appropriate encoding"""
    # create cqt spectrogram    
    spec = CQTSpectrogram(audio_signal=sine_audio, fmin=1, fmax=8000, winstep=0.1, bins_per_octave=32)
    # add annotation
    spec.annotate(labels=(1,2), boxes=((1,2,300,400),(1.5,2.5,300.5,400.5)))
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp12_db.h5')
    h5file = tables.open_file(fpath, 'w')
    # create table
    tbl = di.create_table(h5file=h5file, path='/group_1/', name='table_1', shape=spec.image.shape)
    # write spectrogram to table
    di.write_spec(table=tbl, spec=spec) 
    h5file.close()
    # re-open
    h5file = tables.open_file(fpath, 'r')
    tbl = h5file.get_node("/group_1/table_1")
    # check
    assert int(tbl.attrs.freq_res) == -32
    assert tbl.nrows == 1
    assert tbl[0]['labels'].decode() == '[1,2]'
    assert tbl[0]['boxes'].decode() == '[[1.0,2.0,300.0,400.0],[1.5,2.5,300.5,400.5]]'
    # clean
    h5file.close()
    os.remove(fpath)


@pytest.mark.test_extract
def test_extract(sine_audio):
    """ Test if annotations are correctly extracted from spectrograms"""
    # create spectrogram    
    spec1 = MagSpectrogram(sine_audio, winlen=0.2, winstep=0.02)
    spec1.annotate(labels=(1), boxes=((1.001, 1.401, 50, 300)))
    spec2 = MagSpectrogram(sine_audio, winlen=0.2, winstep=0.02)
    spec2.annotate(labels=(1), boxes=((1.1, 1.5)))
    tshape_orig = spec1.image.shape[0]
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp5_db.h5')
    h5file = tables.open_file(fpath, 'w')
    # create table
    tbl = di.create_table(h5file=h5file, path='/group_1/', name='table_1', shape=spec1.image.shape)
    # write spectrograms to table
    di.write_spec(table=tbl, spec=spec1, id='1')  # Box: 1.0-1.4 s & 50-300 Hz
    di.write_spec(table=tbl, spec=spec2, id='2')  # Box: 1.1-1.5 s Hz
   
    # get segments with label=1
    selection, complements = di.extract(table=tbl, label=1, min_length=0.8, fpad=False, center=True)
    assert len(selection) == 2
    assert len(complements) == 2
    
    assert selection[0].image.shape == (40,51)
    assert complements[0].image.shape[0] == (spec1.image.shape[0] - selection[0].image.shape[0])

    assert selection[1].image.shape == (40,4411)
    assert complements[1].image.shape[0] == (spec2.image.shape[0] - selection[1].image.shape[0])

    tshape = int(0.8 / spec1.tres)
    assert selection[0].image.shape[0] == tshape
    fshape = int(250 / spec1.fres) + 1
    assert selection[0].image.shape[1] == fshape
    assert selection[0].boxes[0][0] == pytest.approx(0.201, abs=0.000001)
    assert selection[0].boxes[0][1] == pytest.approx(0.601, abs=0.000001)

    h5file.close()
    os.remove(fpath)



@pytest.mark.test_parse_labels
def test_parse_labels(sine_audio):
    """Test if labels with the expected format are correctly parsed"""
    # create spectrogram    
    spec1 = MagSpectrogram(sine_audio, winlen=0.2, winstep=0.02)
    spec1.annotate(labels=(1), boxes=[[1.001, 1.401, 50, 300]])
    spec2 = MagSpectrogram(sine_audio, winlen=0.2, winstep=0.02)
    spec2.annotate(labels=(1,2), boxes=[[1.1, 1.5], [1.6, 1.7]])
    tshape_orig = spec1.image.shape[0]
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp6_db.h5')
    h5file = tables.open_file(fpath, 'w')
    # create table
    tbl = di.create_table(h5file=h5file, path='/group_1/', name='table_1', shape=spec1.image.shape)
    # write spectrograms to table
    di.write_spec(table=tbl, spec=spec1, id='1')  # Box: 1.0-1.4 s & 50-300 Hz
    di.write_spec(table=tbl, spec=spec2, id='2')  # Box: 1.1-1.5 s Hz
    #parse labels
    labels = tbl[0]['labels']
    labels = di.parse_labels(labels)
    assert type(labels) == list
    assert labels == [1]

    labels = tbl[1]['labels']
    labels = di.parse_labels(labels)
    assert type(labels) == list
    assert labels == [1,2]

    h5file.close()
    os.remove(fpath)
    

@pytest.mark.test_parse_boxes
def test_parse_boxes(sine_audio):
    """Test if boxes with the expected format are correctly parsed"""
    # create spectrogram    
    spec1 = MagSpectrogram(sine_audio, winlen=0.2, winstep=0.02)
    spec1.annotate(labels=(1), boxes=[[1.001, 1.401, 50, 300]])
    spec2 = MagSpectrogram(sine_audio, winlen=0.2, winstep=0.02)
    spec2.annotate(labels=(1,2), boxes=[[1.1, 1.5], [1.6, 1.7]])
    tshape_orig = spec1.image.shape[0]
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp7_db.h5')
    h5file = tables.open_file(fpath, 'w')
    # create table
    tbl = di.create_table(h5file=h5file, path='/group_1/', name='table_1', shape=spec1.image.shape)
    # write spectrograms to table
    di.write_spec(table=tbl, spec=spec1, id='1')  # Box: 1.0-1.4 s & 50-300 Hz
    di.write_spec(table=tbl, spec=spec2, id='2')  # Box: 1.1-1.5 s Hz
    # parse boxes
    boxes = tbl[0]['boxes']
    boxes = di.parse_boxes(boxes)
    assert type(boxes) == list
    assert boxes == [[1.001, 1.401, 50, 300]]

    boxes = tbl[1]['boxes']
    boxes = di.parse_boxes(boxes)
    assert type(boxes) == list
    assert boxes == [[1.1, 1.5, 0.0, 22050.0], [1.6, 1.7, 0.0, 22050.0]]

    h5file.close()
    os.remove(fpath)


@pytest.mark.filter_by_label
def test_filter_by_label(sine_audio):
    """ Test if filter_by_label works when providing an int or list of ints as the label argument"""
    # create spectrogram  
    spec1 = MagSpectrogram(sine_audio, winlen=0.2, winstep=0.02)
    spec1.annotate(labels=(1), boxes=((1.0, 1.4, 50, 300)))

    spec2 = MagSpectrogram(sine_audio, winlen=0.2, winstep=0.02)
    spec2.annotate(labels=(2), boxes=((1.0, 1.4, 50, 300)))

    spec3 = MagSpectrogram(sine_audio, winlen=0.2, winstep=0.02)
    spec3.annotate(labels=(2,3), boxes=((1.0, 1.4, 50, 300), (2.0, 2.4, 80, 200)))
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp8_db.h5')
    h5file = tables.open_file(fpath, 'w')
    # create table
    tbl = di.create_table(h5file=h5file, path='/group_1/', name='table_1', shape=spec1.image.shape)
    # write spectrogram to table
    di.write_spec(table=tbl, spec=spec1, id='A') 
    di.write_spec(table=tbl, spec=spec1, id='B') 
    di.write_spec(table=tbl, spec=spec2, id='C') 
    di.write_spec(table=tbl, spec=spec2, id='D')
    di.write_spec(table=tbl, spec=spec3, id='E') 
    

    # select spectrograms containing the label 1
    rows = di.filter_by_label(table=tbl, label=1)
    assert len(rows) == 2
    assert rows == [0,1]

    # select spectrograms containing the label 2
    rows = di.filter_by_label(table=tbl, label=[2])
    assert len(rows) == 3
    assert rows == [2,3,4]

    # select spectrograms containing the labels 1 or 3
    rows = di.filter_by_label(table=tbl, label=[1,3])
    assert len(rows) == 3
    assert rows == [0,1,4]

    h5file.close()
    os.remove(fpath)


@pytest.mark.filter_by_label
def test_filter_by_label_raises(sine_audio):
    """ Test if filter_by_label raises expected exception when the the label argument is of the wrong type"""
    # open h5 file
    fpath = os.path.join(path_to_assets, '15x_same_spec.h5')
    h5file = tables.open_file(fpath, 'r')
    tbl = di.open_table(h5file,"/train/species1")
    
    with pytest.raises(TypeError):
        di.filter_by_label(table=tbl, label='a')
    with pytest.raises(TypeError):
        di.filter_by_label(table=tbl, label=['a','b'])
    with pytest.raises(TypeError):        
        di.filter_by_label(table=tbl, label='1')
    with pytest.raises(TypeError):    
        di.filter_by_label(table=tbl, label='1,2')
    with pytest.raises(TypeError):    
        di.filter_by_label(table=tbl, label=b'1')
    with pytest.raises(TypeError):    
        di.filter_by_label(table=tbl, label=1.0)
    with pytest.raises(TypeError):    
        di.filter_by_label(table=tbl, label= (1,2))
    with pytest.raises(TypeError):    
        di.filter_by_label(table=tbl, label=[1.0,2])
   
    h5file.close()



@pytest.mark.load_specs
def test_load_specs_no_index_list():
    """Test if load specs loads the entire table if index_list is None""" 

    fpath = os.path.join(path_to_assets, '15x_same_spec.h5')
    h5file = tables.open_file(fpath, 'r')
    tbl = di.open_table(h5file,"/train/species1")
    
    selected_specs = di.load_specs(tbl)
    assert len(selected_specs) == tbl.nrows
    is_spec = [isinstance(item, Spectrogram) for item in selected_specs]
    assert all(is_spec)
    
    h5file.close()

@pytest.mark.load_specs
def test_load_specs_with_index_list():
    """Test if load_specs loads the spectrograms specified by index_list""" 

    fpath = os.path.join(path_to_assets, '15x_same_spec.h5')
    h5file = tables.open_file(fpath, 'r')
    tbl = di.open_table(h5file,"/train/species1")
    
    selected_specs = di.load_specs(tbl, index_list=[0,3,14])
    assert len(selected_specs) == 3
    is_spec = [isinstance(item, Spectrogram) for item in selected_specs]
    assert all(is_spec)

    h5file.close()

def test_create_spec_database_with_default_args():

    output_file = os.path.join(path_to_assets, 'tmp/db_spec.h5')
    input_dir = os.path.join(path_to_assets, 'wav_files/')

    di.create_spec_database(output_file=output_file, input_dir=input_dir)

    path = os.path.join(path_to_assets, 'tmp/db_spec.h5')
    fil = tables.open_file(path, 'r')
    tbl = di.open_table(fil, "/spec")
    specs = di.load_specs(tbl)
    assert len(specs) == 2
    assert specs[0].tres == 0.02
    tbl = di.open_table(fil, "/subf/spec")
    specs = di.load_specs(tbl)
    assert len(specs) == 1
    assert specs[0].tres == 0.02

    fil.close()

def test_create_spec_database_with_cqt_specs():

    output_file = os.path.join(path_to_assets, 'tmp/db_spec_cqt.h5')
    input_dir = os.path.join(path_to_assets, 'wav_files/')

    di.create_spec_database(output_file=output_file, input_dir=input_dir, cqt=True, bins_per_octave=16)

    path = os.path.join(path_to_assets, 'tmp/db_spec_cqt.h5')
    fil = tables.open_file(path, 'r')
    tbl = di.open_table(fil, "/spec")
    specs = di.load_specs(tbl)
    assert len(specs) == 2
    assert specs[0].bins_per_octave == 16
    tbl = di.open_table(fil, "/subf/spec")
    specs = di.load_specs(tbl)
    assert len(specs) == 1
    assert specs[0].bins_per_octave == 16

    fil.close()

def test_create_spec_database_with_size_limit():

    output_file = os.path.join(path_to_assets, 'tmp/db2_spec.h5')
    input_dir = os.path.join(path_to_assets, 'wav_files/')

    di.create_spec_database(output_file=output_file, input_dir=input_dir, max_size=10E6)

    path = os.path.join(path_to_assets, 'tmp/db2_spec_000.h5')
    fil = tables.open_file(path, 'r')
    tbl = di.open_table(fil, "/spec")
    specs = di.load_specs(tbl)
    assert len(specs) == 1
    tbl = di.open_table(fil, "/subf/spec")
    specs = di.load_specs(tbl)
    assert len(specs) == 1
    fil.close()

    path = os.path.join(path_to_assets, 'tmp/db2_spec_001.h5')
    fil = tables.open_file(path, 'r')
    tbl = di.open_table(fil, "/spec")
    specs = di.load_specs(tbl)
    assert len(specs) == 1
    fil.close()

def test_create_spec_database_with_annotations():

    csvfile = os.path.join(path_to_assets, 'tmp/ann.csv')
    output_file = os.path.join(path_to_assets, 'tmp/db3_spec.h5')
    input_dir = os.path.join(path_to_assets, 'wav_files/')

    # create annotations
    import pandas as pd
    d = {'filename': ['w1.wav'], 'start': [0.1], 'end': [0.7], 'label': [1]}
    df = pd.DataFrame(data=d)
    df.to_csv(csvfile)

    # create database
    di.create_spec_database(output_file=output_file, input_dir=input_dir, annotations_file=csvfile)

    fil = tables.open_file(output_file, 'r')
    tbl = di.open_table(fil, "/spec")
    specs = di.load_specs(tbl)
    assert len(specs) == 2
    assert len(specs[0].labels) == 1
    assert specs[0].labels[0] == 1
    tbl = di.open_table(fil, "/subf/spec")
    specs = di.load_specs(tbl)
    assert len(specs) == 1
    assert len(specs[0].labels) == 0

    fil.close()

    # create database with cqt specs
    output_file = os.path.join(path_to_assets, 'tmp/db3_spec_cqt.h5')
    di.create_spec_database(output_file=output_file, input_dir=input_dir, annotations_file=csvfile, cqt=True)

    fil = tables.open_file(output_file, 'r')
    tbl = di.open_table(fil, "/spec")
    specs = di.load_specs(tbl)
    assert len(specs) == 2
    assert len(specs[0].labels) == 1
    assert specs[0].labels[0] == 1
    tbl = di.open_table(fil, "/subf/spec")
    specs = di.load_specs(tbl)
    assert len(specs) == 1
    assert len(specs[0].labels) == 0

    fil.close()

def test_init_spec_writer():
    out = os.path.join(path_to_assets, 'tmp/db4.h5')
    di.SpecWriter(output_file=out)

def test_spec_writer_can_write_one_spec(sine_audio):
    out = os.path.join(path_to_assets, 'tmp/db5.h5')
    writer = di.SpecWriter(output_file=out)
    spec = MagSpectrogram(sine_audio, 0.5, 0.1)
    writer.write(spec=spec)
    writer.close()
    fname = os.path.join(path_to_assets, 'tmp/db5.h5')
    fil = tables.open_file(fname, 'r')
    assert '/spec' in fil
    specs = di.load_specs(fil.root.spec)
    assert len(specs) == 1

def test_spec_writer_can_write_two_specs_to_same_node(sine_audio):
    out = os.path.join(path_to_assets, 'tmp/db6.h5')
    writer = di.SpecWriter(output_file=out)
    spec = MagSpectrogram(sine_audio, 0.5, 0.1)
    writer.write(spec=spec)
    writer.write(spec=spec)
    writer.close()
    fname = os.path.join(path_to_assets, 'tmp/db6.h5')
    fil = tables.open_file(fname, 'r')
    assert '/spec' in fil
    specs = di.load_specs(fil.root.spec)
    assert len(specs) == 2

def test_spec_writer_can_write_several_specs_to_different_nodes(sine_audio):
    out = os.path.join(path_to_assets, 'tmp/db7.h5')
    writer = di.SpecWriter(output_file=out)
    spec = MagSpectrogram(sine_audio, 0.5, 0.1)
    writer.write(spec=spec, path='/first', name='test')
    writer.write(spec=spec, path='/first', name='test')
    writer.write(spec=spec, path='/second', name='temp')
    writer.write(spec=spec, path='/second', name='temp')
    writer.write(spec=spec, path='/second', name='temp')
    writer.close()
    fname = os.path.join(path_to_assets, 'tmp/db7.h5')
    fil = tables.open_file(fname, 'r')
    assert '/first/test' in fil
    assert '/second/temp' in fil
    specs = di.load_specs(fil.root.first.test)
    assert len(specs) == 2
    specs = di.load_specs(fil.root.second.temp)
    assert len(specs) == 3

def test_spec_writer_splits_into_several_files_when_max_size_is_reached(sine_audio):
    out = os.path.join(path_to_assets, 'tmp/db8.h5')
    writer = di.SpecWriter(output_file=out, max_size=1E6) # max size: 1 Mbyte
    spec = MagSpectrogram(sine_audio, 0.5, 0.1)
    writer.write(spec=spec)
    writer.write(spec=spec)
    writer.write(spec=spec)
    writer.close()

    fname = os.path.join(path_to_assets, 'tmp/db8_000.h5')
    fil = tables.open_file(fname, 'r')
    assert '/spec' in fil
    specs = di.load_specs(fil.root.spec)
    assert len(specs) == 2

    fname = os.path.join(path_to_assets, 'tmp/db8_001.h5')
    fil = tables.open_file(fname, 'r')
    assert '/spec' in fil
    specs = di.load_specs(fil.root.spec)
    assert len(specs) == 1

def test_spec_writer_change_directory(sine_audio):
    out = os.path.join(path_to_assets, 'tmp/db9.h5')
    writer = di.SpecWriter(output_file=out)
    spec = MagSpectrogram(sine_audio, 0.5, 0.1)
    writer.cd('/home/fish')
    writer.write(spec=spec)
    writer.write(spec=spec)
    writer.write(spec=spec)
    writer.cd('/home/whale')
    writer.write(spec=spec)
    writer.write(spec=spec)
    writer.close()
    fname = os.path.join(path_to_assets, 'tmp/db9.h5')
    fil = tables.open_file(fname, 'r')
    assert '/home/fish' in fil
    assert '/home/whale' in fil
    specs = di.load_specs(fil.root.home.fish)
    assert len(specs) == 3
    specs = di.load_specs(fil.root.home.whale)
    assert len(specs) == 2
    fil.close()

def test_two_spec_writers_simultaneously(sine_audio):
    # init two spec writers
    out1 = os.path.join(path_to_assets, 'tmp/db10.h5')
    writer1 = di.SpecWriter(output_file=out1)
    out2 = os.path.join(path_to_assets, 'tmp/db11.h5')
    writer2 = di.SpecWriter(output_file=out2)
    # create spec
    spec = MagSpectrogram(sine_audio, 0.5, 0.1)
    # write 
    writer1.cd('/home/fish')
    writer1.write(spec=spec)
    writer1.write(spec=spec)
    writer1.write(spec=spec)
    writer2.cd('/home/whale')
    writer2.write(spec=spec)
    writer2.write(spec=spec)
    # close
    writer1.close()
    writer2.close()
    # check file 1
    fil1 = tables.open_file(out1, 'r')
    assert '/home/fish' in fil1
    specs = di.load_specs(fil1.root.home.fish)
    assert len(specs) == 3
    fil1.close()
    # check file 2
    fil2 = tables.open_file(out2, 'r')
    assert '/home/whale' in fil2
    specs = di.load_specs(fil2.root.home.whale)
    assert len(specs) == 2
    fil2.close()


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Parsing module within the ketos library

    This module provides utilities to parse various string 
    structures.

    Contents:
        WinFun class: 
"""

import json
from collections import namedtuple
from pint import UnitRegistry # SI units
from enum import Enum


class WinFun(Enum):
    HAMMING = 1


ureg = UnitRegistry()


Interval = namedtuple('Interval', 'low high')
Interval.__doc__ = '''\
Numerical intervals

low - Lower limit (float)
high - Upper limit (float)''' 


SpectrogramConfiguration = namedtuple('SpectrogramConfiguration', 'rate window_size step_size window_function low_frequency_cut high_frequency_cut')
SpectrogramConfiguration.__doc__ = '''\
Configuration parameters for generation of spectrograms

rate - Sampling rate in Hz (int)
window_size - Window size used for framing in seconds (float)
step_size - Step size used for framing in seconds (float)
window_function - Window function used for framing (e.g. Hamming window)
low_frequency_cut - Low-frequency cut-off in Hz (float)
high_frequency_cut - High-frequency cut-off in Hz (float)''' 


def parse_spectrogram_configuration(data):
    """ Parse configuration settings for generating spectrograms.

        Any setting not specified in the json string will be set 
        to None.

    Args:
        data : str
            Json-format string with the configuration settings 
    
    Returns:
        c : SpectrogramConfiguration
            Spectrogram configuration settings

    Example:
    
        >>> import json
        >>> import ketos.data_handling.parsing as par
        >>> 
        >>> input = '{"spectrogram": {"rate": "20 kHz", "window_size": "0.1 s", "step_size": "0.025 s", "window_function": "HAMMING", "low_frequency_cut": "30Hz", "high_frequency_cut": "3000Hz"}}'
        >>> data = json.loads(input)
        >>> settings = par.parse_spectrogram_configuration(data['spectrogram'])
        >>> print(settings.rate)  # print sampling rate in Hz
        20000.0
    """
    Q = ureg.Quantity

    # default values
    rate, wsiz, step, wfun, flow, fhigh = None, None, None, None, None, None    

    # check that entry exists, before attempting to read
    if data.get('rate') is not None:
        rate = Q(data['rate'])
        rate = rate.m_as("Hz")

    if data.get('window_size') is not None:
        wsiz = Q(data['window_size'])
        wsiz = wsiz.m_as("s")

    if data.get('step_size') is not None:
        step = Q(data['step_size'])
        step = step.m_as("s")

    if data.get('window_function') is not None:

        for name, member in WinFun.__members__.items():
            if data['window_function'] == name:
                wfun = member

        if wfun is None:
            s = ", ".join(name for name, _ in WinFun.__members__.items())
            raise ValueError("Unknown window function. Select between: "+s)

    if data.get('low_frequency_cut') is not None:
        flow = Q(data['low_frequency_cut'])
        flow = flow.m_as("Hz")

    if data.get('high_frequency_cut') is not None:
        fhigh = Q(data['high_frequency_cut'])
        fhigh = fhigh.m_as("Hz")

    c = SpectrogramConfiguration(rate, wsiz, step, wfun, flow, fhigh)    

    return c


def parse_frequency_bands(data):
    """ Parse list of frequency bands

    Args:
        data : str
            Json-format string frequency bands 
    
    Returns:
        name : list(str)
            Band names
        freq_intv: list(Interval)
            Band frequency ranges in Hz 

    Example:
    
        >>> import json
        >>> import ketos.data_handling.parsing as par
        >>> 
        >>> input = '{"frequency_bands": [{"name": "A", "range": ["11.0Hz", "22.1Hz"]},{"name": "B", "range": ["9kHz", "10kHz"]}]}'
        >>> data = json.loads(input)
        >>> names, bands = par.parse_frequency_bands(data['frequency_bands'])
        >>> print(names)  # print names of frequency bands
        ['A', 'B']
        >>> for b in bands: 
        ...     print(b.low, b.high) # print frequency range
        11.0 22.1
        9000.0 10000.0
    """
    Q = ureg.Quantity
    
    name, freq_intv = list(), list()

    for band in data:

        assert band.get('name') is not None, 'name field is required for frequency band'
        assert band.get('range') is not None, 'range field is required for frequency band'

        name.append(band['name'])

        low = Q(band['range'][0])
        low = low.m_as('Hz')

        high = Q(band['range'][1])
        high = high.m_as('Hz')

        intv = Interval(low, high)
        freq_intv.append(intv)

    return name, freq_intv


def str2bool(v):
    """ Convert most common answers to yes/no questions to boolean

    Args:
        v : str
            Answer 
    
    Returns:
        res : bool
            Answer converted to boolean 
    """
    res = v.lower() in ("yes", "YES", "Yes", "true", "True", "TRUE", "on", "ON", "t", "T", "1")
    return res




# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'annotation' module within the ketos library
"""

import pytest
import numpy as np
from ketos.audio_processing.annotation import AnnotationHandler


def test_can_initialize_empty_annotation_handler():
    _ = AnnotationHandler()

def test_can_initialize_annotation_handler():
    # one label, one 4d box
    _ = AnnotationHandler(labels=[1], boxes=[[0.1, 0.3, 100., 400.]])
    # one label, one 2d box
    _ = AnnotationHandler(labels=[1], boxes=[[0.1, 0.3]])
    # one label, two boxes
    with pytest.raises(AssertionError):
        _ = AnnotationHandler(labels=[1], boxes=[[0.1, 0.3],[0.1, 0.3]])
    # two labels, one box
    with pytest.raises(AssertionError):
        _ = AnnotationHandler(labels=[1,2], boxes=[[0.1, 0.3]])
    # one label, one box
    _ = AnnotationHandler(labels=1, boxes=[0.1, 0.3, 100., 400.])
    # box with wrong dimensions
    with pytest.raises(AssertionError):
        _ = AnnotationHandler(labels=1, boxes=[0.1, 0.3, 100])
        _ = AnnotationHandler(labels=1, boxes=[0.1, 0.3, 100, 333, 1])
    # labels provided as tuples
    _ = AnnotationHandler(labels=(1,2), boxes=[[0.1, 0.3],[0.4,0.7]])
    # labels provided as numpy arrays
    _ = AnnotationHandler(labels=np.array([1,2]), boxes=[[0.1, 0.3],[0.4,0.7]])

def test_add_annotations():
    a = AnnotationHandler()
    a.annotate(labels=1, boxes=[1,2,3,4])
    a.annotate(labels=[2,3], boxes=[[1,2,3,4],[5,6,7,8]])
    assert len(a.labels) == 3
    # two labels, and two boxes
    a = AnnotationHandler()
    labels = [0, 1]
    boxes = [[10.0, 12.2, 110., 700.],[30., 34.]]
    a.annotate(labels, boxes)

def test_delete_annotations():
    labels = [1,2,3]
    boxes = [[1,2,3,4],[5,6,7,8],[9,10,11,12]]
    # delete one annotation
    a = AnnotationHandler(labels, boxes)
    assert len(a.labels) == 3
    a.delete_annotations(id=0)
    assert len(a.labels) == 2
    assert a.labels[0] == 2
    # delete two annotations
    a = AnnotationHandler(labels, boxes)
    assert len(a.labels) == 3
    a.delete_annotations(id=[0,1])
    assert len(a.labels) == 1
    assert a.labels[0] == 3
    # delete a non-existing annotation
    a = AnnotationHandler(labels, boxes)
    assert len(a.labels) == 3
    a.delete_annotations(id=77)
    assert len(a.labels) == 3
    assert a.labels[0] == 1
    # delete all annotation
    a = AnnotationHandler(labels, boxes)
    assert len(a.labels) == 3
    a.delete_annotations()
    assert len(a.labels) == 0

def test_get_cropped_annotations():
    labels = [1,2]
    box1 = [14.,17.,0.,29.]
    box2 = [2.1,13.0,1.1,28.5]
    boxes = [box1, box2]
    a = AnnotationHandler(labels=labels, boxes=boxes)
    l, b = a.get_cropped_annotations(t1=2, t2=15, f1=24, f2=27)
    assert len(l) == 2
    assert l[0] == 1
    assert b[0][0] == pytest.approx(14.0, abs=0.001)
    assert b[0][1] == pytest.approx(15.0, abs=0.001)
    assert b[0][2] == pytest.approx(24.0, abs=0.001)
    assert b[0][3] == pytest.approx(27.0, abs=0.001)
    assert l[1] == 2
    assert b[1][0] == pytest.approx(2.1, abs=0.001)
    assert b[1][1] == pytest.approx(13.0, abs=0.001)
    assert b[1][2] == pytest.approx(24.0, abs=0.001)
    assert b[1][3] == pytest.approx(27.0, abs=0.001)

def test_get_cropped_annotations_2():
    labels = [0, 1]
    boxes = [[10.0, 12.2, 110., 700.],[30., 34.]]
    handler = AnnotationHandler(labels, boxes)
    l, b = handler.get_cropped_annotations(t1=10.3, f2=555.)
    assert len(l) == 2
    assert l[0] == 0
    assert l[1] == 1
    assert len(b) == 2
    assert b[0][0] == 10.3
    assert b[0][1] == 12.2
    assert b[0][2] == 110.0
    assert b[0][3] == 555.0
    assert b[1][0] == 30.
    assert b[1][1] == 34.
    assert b[1][2] == 0
    assert b[1][3] == 555.0

def test_shift_annotations():
    labels = [1,2]
    box1 = [14.,17.,0.,29.]
    box2 = [2.1,13.0,1.1,28.5]
    boxes = [box1, box2]
    a = AnnotationHandler(labels=labels, boxes=boxes)
    a._shift_annotations(delay=3.3)
    l = a.labels
    b = a.boxes
    assert len(l) == 2
    assert b[0][0] == pytest.approx(17.3, abs=0.001)
    assert b[0][1] == pytest.approx(20.3, abs=0.001)
    assert b[1][0] == pytest.approx(5.4, abs=0.001)
    assert b[1][1] == pytest.approx(16.3, abs=0.001)



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'data_handling' module within the ketos library
"""

import pytest
import numpy as np
import pandas as pd
import ketos.data_handling.data_handling as dh
import ketos.audio_processing.audio_processing as ap
from ketos.audio_processing.spectrogram import MagSpectrogram
from ketos.data_handling.data_handling import AudioSequenceReader, AnnotationTableReader
from ketos.audio_processing.audio import AudioSignal
import datetime
import shutil
import os
from glob import glob

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


today = datetime.datetime.today()



@pytest.mark.parametrize("input,depth,expected",[
    (1,2,np.array([0,1])),
    (0,2,np.array([1,0])),
    (1.0,2,np.array([0,1])),
    (0.0,2,np.array([1,0])),
    ])
@pytest.mark.test_to1hot
def test_to1hot_works_with_floats_and_ints(input, depth, expected):
    one_hot = dh.to1hot(input, depth)
    assert (one_hot == expected).all()


@pytest.mark.parametrize("input,depth,expected",[
    (1,4,np.array([0,1,0,0])),
    (1,4,np.array([0,1,0,0])),
    (1,2,np.array([0,1])),
    (1,10,np.array([0,1,0,0,0,0,0,0,0,0])),
    ])
@pytest.mark.test_to1hot
def test_to1hot_output_has_correct_depth(input,depth, expected):
    one_hot = dh.to1hot(input,depth)
    assert len(one_hot) == depth


@pytest.mark.parametrize("input,depth,expected",[
    (3,4,np.array([0,0,0,1])),
    (0,4,np.array([1,0,0,0])),
    (1.0,2,np.array([0,1])),
    (5.0,10,np.array([0,0,0,0,0,1,0,0,0,0])),
    ])
@pytest.mark.test_to1hot
def test_to1hot_works_with_multiple_categories(input,depth, expected):
    one_hot = dh.to1hot(input,depth)
    assert (one_hot == expected).all()


@pytest.mark.parametrize("input,depth,expected",[
    (np.array([3,0,1,5]),6,
     np.array([[0., 0., 0., 1., 0., 0.],
              [1., 0., 0., 0., 0., 0.],
              [0., 1., 0., 0., 0., 0.],
              [0., 0., 0., 0., 0., 1.]])),
    (np.array([0,1]),3,
     np.array([[1., 0., 0.],
               [0., 1., 0.]])),
    ])
@pytest.mark.test_to1hot
def test_to1hot_works_with_multiple_input_values_at_once(input,depth, expected):
    one_hot = dh.to1hot(input,depth)
    assert (one_hot == expected).all()


@pytest.mark.parametrize("input,depth,expected",[
    (pd.DataFrame({"label":[0,0,1,0,1,0]}),2,
     np.array([[1.0, 0.0],
                [1.0, 0.0],
                [0.0, 1.0],
                [1.0, 0.0],
                [0.0, 1.0],
                [1.0, 0.0]]),)
    ])
@pytest.mark.test_to1hot
def test_to1hot_works_when_when_applying_to_DataFrame(input,depth, expected):
     
    one_hot = input["label"].apply(dh.to1hot,depth=depth)
    for i in range(len(one_hot)):
        assert (one_hot[i] == expected[i]).all()

@pytest.mark.test_find_wave_files
def test_find_wave_files():
    dir = os.path.join(path_to_assets,'test_find_wave_files')
    #delete directory and files within
    if os.path.exists(dir):
        shutil.rmtree(dir)
    dh.create_dir(dir)
    # create two wave files
    f1 = os.path.join(dir, "f1.wav")
    f2 = os.path.join(dir, "f2.wav")
    ap.wave.write(f2, rate=100, data=np.array([1.,0.]))
    ap.wave.write(f1, rate=100, data=np.array([0.,1.]))
    # get file names
    files = dh.find_wave_files(dir, fullpath=False)
    assert len(files) == 2
    assert files[0] == "f1.wav"
    assert files[1] == "f2.wav"
    files = dh.find_wave_files(dir, fullpath=True)
    assert len(files) == 2
    assert files[0] == f1
    assert files[1] == f2
    #delete directory and files within
    shutil.rmtree(dir)

def test_find_wave_files_from_multiple_folders():
    folder = path_to_assets + "/sub"
    # create two wave files in separate subfolders
    sub1 = folder + "/sub1"
    sub2 = folder + "/sub2"
    if not os.path.exists(sub1):
        os.makedirs(sub1)
    if not os.path.exists(sub2):
        os.makedirs(sub2)
    # clean
    for f in glob(sub1 + "/*.wav"):
        os.remove(f)  #clean
    for f in glob(sub2 + "/*.wav"):
        os.remove(f)  #clean
    f1 = sub1 + "/f1.wav"
    f2 = sub2 + "/f2.wav"
    ap.wave.write(f2, rate=100, data=np.array([1.,0.]))
    ap.wave.write(f1, rate=100, data=np.array([0.,1.]))
    # get file names
    files = dh.find_wave_files(folder, fullpath=False, subdirs=True)
    assert len(files) == 2
    assert files[0] == "f1.wav"
    assert files[1] == "f2.wav"
    files = dh.find_wave_files(folder, fullpath=True, subdirs=True)
    assert len(files) == 2
    assert files[0] == f1
    assert files[1] == f2

    
################################
# from1hot() tests
################################


@pytest.mark.parametrize("input,expected",[
    (np.array([0,1]),1),
    (np.array([1,0]),0),
    (np.array([0.0,1.0]),1),
    (np.array([1.0,0.0]),0),
    ])
@pytest.mark.test_from1hot
def test_from1hot_works_with_floats_and_ints(input, expected):
    one_hot = dh.from1hot(input)
    assert one_hot == expected


@pytest.mark.parametrize("input,expected",[
    (np.array([0,0,0,1]),3),
    (np.array([1,0,0,0]),0),
    (np.array([0,1]),1),
    (np.array([0,0,0,0,0,1,0,0,0,0]),5),
    ])
@pytest.mark.test_from1hot
def test_from1hot_works_with_multiple_categories(input, expected):
    one_hot = dh.from1hot(input)
    assert one_hot == expected


@pytest.mark.parametrize("input,expected",[
    (np.array([[0., 0., 0., 1., 0., 0.],
              [1., 0., 0., 0., 0., 0.],
              [0., 1., 0., 0., 0., 0.],
              [0., 0., 0., 0., 0., 1.]]),np.array([3,0,1,5])),
    (np.array([[1., 0., 0.],
               [0., 1., 0.]]), np.array([0,1])),
    ])
@pytest.mark.test_from1hot
def test_from1hot_works_with_multiple_input_values_at_once(input, expected):
    one_hot = dh.from1hot(input)
    assert (one_hot == expected).all()

@pytest.mark.test_read_wave
def test_read_wave_file(sine_wave_file):
    rate, data = dh.read_wave(sine_wave_file)
    assert rate == 44100

@pytest.mark.test_parse_datetime
def test_parse_datetime_with_urban_sharks_format():
    fname = 'empty_HMS_12_ 5_28__DMY_23_ 2_84.wav'
    full_path = os.path.join(path_to_assets, fname)
    ap.wave.write(full_path, rate=1000, data=np.array([0.]))
    fmt = '*HMS_%H_%M_%S__DMY_%d_%m_%y*'
    dt = dh.parse_datetime(to_parse=fname, fmt=fmt)
    os.remove(full_path)
    assert dt is not None
    assert dt.year == 2084
    assert dt.month == 2
    assert dt.day == 23
    assert dt.hour == 12
    assert dt.minute == 5
    assert dt.second == 28

@pytest.mark.test_parse_datetime
def test_parse_datetime_with_non_matching_format():
    fname = 'empty_HMQ_12_ 5_28__DMY_23_ 2_84.wav'
    full_path = os.path.join(path_to_assets, fname)
    ap.wave.write(full_path, rate=1000, data=np.array([0.]))
    fmt = '*HMS_%H_%M_%S__DMY_%d_%m_%y*'
    dt = dh.parse_datetime(to_parse=fname, fmt=fmt)
    os.remove(full_path)
    assert dt == None


@pytest.mark.parse_seg_name
def test_parse_seg_name():
    id,labels = dh.parse_seg_name('id_rb001_89_l_[0].wav')
    assert id == 'rb001_89'
    assert labels == '[0]' 

    id,labels = dh.parse_seg_name('id_rb001_89_l_[0]')
    assert id == 'rb001_89'
    assert labels == '[0]' 

    id,labels = dh.parse_seg_name('id_rb001_89_l_[1,2].wav')
    assert id == 'rb001_89'
    assert labels == '[1,2]' 

    id,labels = dh.parse_seg_name('id_rb001_89_l_[1,2]')
    assert id == 'rb001_89'
    assert labels == '[1,2]' 


@pytest.mark.test_divide_audio_into_segments
def test_creates_correct_number_of_segments():
    audio_file = path_to_assets + "/2min.wav"
    annotations = pd.DataFrame({'filename':['2min.wav','2min.wav','2min.wav'],
                                 'label':[1,2,1], 'start':[5.0, 70.34, 105.8],
                                 'end':[6.0,75.98,110.0]})

    try:
        shutil.rmtree(path_to_tmp + "/2s_segs")
    except FileNotFoundError:
        pass

    dh.divide_audio_into_segs(audio_file=audio_file,
        seg_duration=2.0, annotations=annotations, save_to=path_to_tmp + "/2s_segs")
    
    n_seg = len(glob(path_to_tmp + "/2s_segs/id_2min*.wav"))
    assert n_seg == 60



    shutil.rmtree(path_to_tmp + "/2s_segs")


@pytest.mark.test_divide_audio_into_segments
def test_start_end_args():
    audio_file = path_to_assets+ "/2min.wav"
    _= pd.DataFrame({'filename':['2min.wav','2min.wav','2min.wav'],
                                 'label':[1,2,1], 'start':[5.0, 70.34, 105.8],
                                 'end':[6.0,75.98,110.0]})

    try:
        shutil.rmtree(path_to_tmp + "/2s_segs")
    except FileNotFoundError:
        pass

    dh.divide_audio_into_segs(audio_file=audio_file,
        seg_duration=2.0, start_seg=10, end_seg=19, save_to=path_to_tmp + "/2s_segs")
    
    n_seg = len(glob(path_to_tmp + "/2s_segs/id_2min*.wav"))
    assert n_seg == 10



    shutil.rmtree(path_to_tmp + "/2s_segs")

@pytest.mark.test_divide_audio_into_segments
def test_seg_labels_are_correct():
    audio_file = path_to_assets+ "/2min.wav"
    annotations = pd.DataFrame({'filename':['2min.wav','2min.wav','2min.wav'],
                                 'label':[1,2,1], 'start':[5.0, 70.5, 105.0],
                                 'end':[6.0,73.0,108.0]})

    try:
        shutil.rmtree(path_to_tmp + "/2s_segs")
    except FileNotFoundError:
        pass

    dh.divide_audio_into_segs(audio_file=audio_file,
        seg_duration=2.0, annotations=annotations, save_to=path_to_tmp + "/2s_segs")
    
    label_0 = len(glob(path_to_tmp + "/2s_segs/id_2min*l_[[]0].wav"))
    assert label_0 == 53

    label_1 = len(glob(path_to_tmp + "/2s_segs/id_2min*l_[[]1].wav"))
    assert label_1 == 5

    label_2 = len(glob(path_to_tmp + "/2s_segs/id_2min*l_[[]2].wav"))
    assert label_2 == 2

    shutil.rmtree(path_to_tmp + "/2s_segs")


@pytest.mark.test_divide_audio_into_segments
def test_creates_segments_without_annotations():
    audio_file = path_to_assets+ "/2min.wav"
    
    try:
        shutil.rmtree(path_to_tmp + "/2s_segs")
    except FileNotFoundError:
        pass

    dh.divide_audio_into_segs(audio_file=audio_file,
        seg_duration=2.0, annotations=None, save_to=path_to_tmp + "/2s_segs")
    
    n_seg = len(glob(path_to_tmp + "/2s_segs/id_2min*l_[[]NULL].wav"))

    assert n_seg == 60
    shutil.rmtree(path_to_tmp + "/2s_segs")


@pytest.mark.test_seg_from_time_tag
def test_seg_from_time_tag():

    
    audio_file = os.path.join(path_to_assets, "2min.wav")
    
    try:
        shutil.rmtree(os.path.join(path_to_tmp, "from_tags"))
    except FileNotFoundError:
        pass

    dh.create_dir(os.path.join(path_to_tmp, "from_tags"))
    
    dh.seg_from_time_tag(audio_file=audio_file, start=0.5, end=2.5 , name="seg_1.wav", save_to=os.path.join(path_to_tmp, "from_tags") )

    
    rate, sig  = ap.wave.read(os.path.join(path_to_tmp, "from_tags", "seg_1.wav"))
    duration = len(sig)/rate
    assert duration == 2.0
    shutil.rmtree(os.path.join(path_to_tmp, "from_tags"))

@pytest.mark.test_seg_from_annotations
def test_segs_from_annotations():
    audio_file_path = os.path.join(path_to_assets,'2min.wav')
    annotations = pd.DataFrame({'filename':[audio_file_path,audio_file_path,audio_file_path],
                                 'label':[1,2,1], 'start':[5.0, 70.5, 105.0],
                                 'end':[6.0,73.0,108.0]})

    try:
        shutil.rmtree(path_to_assets + "/from_annot")
    except FileNotFoundError:
        pass
    dh.segs_from_annotations(annotations,path_to_assets + "/from_annot")
    
    # label_0 = len(glob(path_to_assets + "/from_annot/id_2min*l_[[]0].wav"))
    # assert label_0 == 53

    label_1 = len(glob(path_to_assets + "/from_annot/id_2min*l_[[]1].wav"))
    assert label_1 == 2

    label_2 = len(glob(path_to_assets + "/from_annot/id_2min*l_[[]2].wav"))
    assert label_2 == 1

    shutil.rmtree(path_to_assets + "/from_annot")
    


@pytest.mark.parametrize("start,end,expected_label",[
    (4.0,5.0,'[1]'),
    (4.0,5.5,'[1]'),
    (5.0,6.0,'[1]'),
    (5.1,6.0,'[1]'),
    (100.0,100.5,'[2]'),
    (100.5,101.0,'[2]'),
    (99.0,103.0,'[2]'),
    (90.0,110.0,'[2, 1]'),
     ])
@pytest.mark.test_get_labels
def test_get_correct_labels(start,end,expected_label):
    audio_file="2min"
    annotations = pd.DataFrame({'filename':['2min.wav','2min.wav','2min.wav'],
                                 'label':[1,2,1], 'start':[5.0, 100.5, 105.0],
                                 'end':[6.0,103.0,108.0]})
    
    label = dh.get_labels(file='2min',start=start, end=end,
                             annotations=annotations, not_in_annotations=0)
    print(label)
    assert label == expected_label
    
@pytest.mark.test_filter_annotations_by_filename
def test_filter_annotations_by_filename():
     annotations = pd.DataFrame({'filename':['2min_01.wav','2min_01.wav','2min_02.wav','2min_02.wav','2min_02.wav'],
                                 'label':[1,2,1,1,1], 'start':[5.0, 100.5, 105.0, 80.0, 90.0],
                                 'end':[6.0,103.0,108.0, 87.0, 94.0]})

     annot_01 = dh._filter_annotations_by_filename(annotations,'2min_01')
     assert annot_01.equals(pd.DataFrame({'filename':['2min_01.wav','2min_01.wav'],
                                 'label':[1,2], 'start':[5.0, 100.5],
                                 'end':[6.0,103.0]}))
                                 
     annot_02 = dh._filter_annotations_by_filename(annotations,'2min_02')
     assert annot_02.equals(pd.DataFrame({'filename':['2min_02.wav','2min_02.wav','2min_02.wav'],
                                 'label':[1,1,1], 'start':[105.0, 80.0, 90.0],
                                 'end':[108.0, 87.0, 94.0]}, index=[2,3,4]))
 
     annot_03 = dh._filter_annotations_by_filename(annotations,'2min_03')               
     assert annot_03.empty
 
@pytest.mark.test_pad_signal
def test_pad_signal():

    sig=np.ones((100))
    rate = 50 
    desired_length = 3.0
    sig_length = len(sig)/rate #sig length in seconds

    padded = dh.pad_signal(signal=sig, rate=rate, length=desired_length)
    

    assert len(padded) == desired_length * rate
    pad_1_limit = int((desired_length - sig_length) * rate / 2)
    pad_2_limit = int(desired_length * rate - pad_1_limit)
    assert sum(padded[:pad_1_limit]) == 0
    assert sum(padded[pad_2_limit:]) == 0
    assert pytest.approx(padded[pad_1_limit:pad_2_limit], sig)

    


def test_init_batch_reader_with_single_file(sine_wave_file):
    reader = AudioSequenceReader(source=sine_wave_file)
    assert len(reader.files) == 1
    assert reader.files[0][0] == sine_wave_file

def test_init_batch_reader_with_two_files(sine_wave_file, sawtooth_wave_file):
    reader = AudioSequenceReader(source=[sine_wave_file, sawtooth_wave_file])
    assert len(reader.files) == 2
    assert reader.files[0][0] == sine_wave_file
    assert reader.files[1][0] == sawtooth_wave_file

@pytest.fixture
def five_time_stamped_wave_files():

    files = list()
    N = 5

    folder = path_to_tmp + '/five_time_stamped_wave_files/'
    if not os.path.exists(folder):
        os.makedirs(folder)

    for i in range(N):
        fname = 'empty_HMS_12_ 5_ {0}__DMY_23_ 2_84.wav'.format(i)
        full_path = os.path.join(folder, fname)
        a = AudioSignal(rate=1000, data=np.zeros(500))
        a.to_wav(full_path)
        files.append(full_path)

    yield folder

    for f in files:
        os.remove(f)

def test_init_batch_reader_with_directory(five_time_stamped_wave_files):
    folder = five_time_stamped_wave_files
    reader = AudioSequenceReader(source=folder)
    assert len(reader.files) == 5

def test_batch_reader_can_parse_date_time(five_time_stamped_wave_files):
    folder = five_time_stamped_wave_files
    print(folder)
    fmt = '*HMS_%H_%M_%S__DMY_%d_%m_%y*'
    reader = AudioSequenceReader(source=folder, datetime_fmt=fmt)
    b = reader.next(700)
    assert b.begin() == datetime.datetime(year=2084, month=2, day=23, hour=12, minute=5, second=0, microsecond=0)
    b = reader.next(600)
    assert b.begin() == datetime.datetime(year=2084, month=2, day=23, hour=12, minute=5, second=1, microsecond=0)
    b = reader.next(300)
    assert b.begin() == datetime.datetime(year=2084, month=2, day=23, hour=12, minute=5, second=2, microsecond=0)
    b = reader.next()
    assert b.begin() == datetime.datetime(year=2084, month=2, day=23, hour=12, minute=5, second=2, microsecond=int(3E5))

def test_batch_reader_log_has_correct_data(five_time_stamped_wave_files):
    folder = five_time_stamped_wave_files
    fmt = '*HMS_%H_%M_%S__DMY_%d_%m_%y*'
    reader = AudioSequenceReader(source=folder, datetime_fmt=fmt)
    reader.next()
    log = reader.log()
    for i in range(5):
        assert log['time'][i] == datetime.datetime(year=2084, month=2, day=23, hour=12, minute=5, second=i, microsecond=0)
        fname = 'empty_HMS_12_ 5_ {0}__DMY_23_ 2_84.wav'.format(i)
        full_path = os.path.join(folder, fname)
        assert log['file'][i] == full_path
    reader.reset()
    b = reader.next(700)
    b = reader.next(600)
    b = reader.next(300)
    b = reader.next()
    for i in range(5):
        assert log['time'][i] == datetime.datetime(year=2084, month=2, day=23, hour=12, minute=5, second=i, microsecond=0)
        fname = 'empty_HMS_12_ 5_ {0}__DMY_23_ 2_84.wav'.format(i)
        full_path = os.path.join(folder, fname)
        assert log['file'][i] == full_path


def test_next_batch_with_single_file(sine_wave_file):
    s = AudioSignal.from_wav(sine_wave_file)
    reader = AudioSequenceReader(source=sine_wave_file)
    assert reader.finished() == False
    b = reader.next()
    assert reader.finished() == True
    assert b.duration() == s.duration()

def test_next_batch_with_multiple_files(sine_wave_file, sawtooth_wave_file, const_wave_file):
    reader = AudioSequenceReader(source=[sine_wave_file, sawtooth_wave_file, const_wave_file])
    b = reader.next()
    s1 = AudioSignal.from_wav(sine_wave_file)
    s2 = AudioSignal.from_wav(sawtooth_wave_file)
    s3 = AudioSignal.from_wav(const_wave_file)
    assert len(b.data) == len(s1.data) + len(s2.data) + len(s3.data) - 2 * reader.n_smooth
    assert reader.finished() == True

def test_next_batch_with_two_files_and_limited_batch_size(sine_wave_file, sawtooth_wave_file):
    s1 = AudioSignal.from_wav(sine_wave_file)
    s2 = AudioSignal.from_wav(sawtooth_wave_file)
    n1 = len(s1.data)
    n2 = len(s2.data)
    size = int((n1+n2) / 1.5)
    reader = AudioSequenceReader(source=[sine_wave_file, sawtooth_wave_file])
    b = reader.next(size)
    assert reader.finished() == False
    assert len(b.data) == size
    b = reader.next(size)
    assert reader.finished() == True
    assert len(b.data) == n1 + n2 - reader.n_smooth - size

def test_next_batch_with_two_very_short_files():
    short_file_1 = os.path.join(path_to_assets, "super_short_1.wav")
    ap.wave.write(short_file_1, rate=4000, data=np.array([1.,2.,3.]))
    short_file_2 = os.path.join(path_to_assets, "super_short_2.wav")
    ap.wave.write(short_file_2, rate=4000, data=np.array([1.,2.,3.]))
    s1 = AudioSignal.from_wav(short_file_1)
    s2 = AudioSignal.from_wav(short_file_2)
    n1 = len(s1.data)
    n2 = len(s2.data)
    reader = AudioSequenceReader(source=[short_file_1, short_file_2])
    b = reader.next()
    assert reader.finished() == True
    assert len(b.data) == n1+n2-2

def test_next_batch_with_empty_file_and_resampling():
    empty = os.path.join(path_to_assets, "empty.wav")
    ap.wave.write(empty, rate=4000, data=np.empty(0))
    s = AudioSignal.from_wav(empty)
    n = len(s.data)
    reader = AudioSequenceReader(source=[empty], rate=2000)
    b = reader.next()
    assert reader.finished() == True
    assert b is None

def test_next_batch_with_multiple_files(sine_wave_file, sawtooth_wave_file, const_wave_file):
    reader = AudioSequenceReader(source=[sine_wave_file, sawtooth_wave_file, const_wave_file])
    b = reader.next()
    s1 = AudioSignal.from_wav(sine_wave_file)
    s2 = AudioSignal.from_wav(sawtooth_wave_file)
    s3 = AudioSignal.from_wav(const_wave_file)
    assert len(b.data) == len(s1.data) + len(s2.data) + len(s3.data) - 2 * reader.n_smooth
    assert reader.finished() == True

def test_init_annotation_table_reader():
    fname = os.path.join(path_to_assets, 'dummy_annotations.csv')
    AnnotationTableReader(fname)
    fname = os.path.join(path_to_assets, 'dummy_annotations_w_freq.csv')
    AnnotationTableReader(fname)

def test_get_annotations_for_file_with_one_annotation():
    fname = os.path.join(path_to_assets, 'dummy_annotations.csv')
    a = AnnotationTableReader(fname)
    l, b = a.get_annotations('x.wav')
    assert len(l) == 1
    assert l[0] == 0
    assert len(b) == 1
    assert b[0][0] == 1
    assert b[0][1] == 2
    assert b[0][2] == 0
    import math
    assert b[0][3] == math.inf
    fname = os.path.join(path_to_assets, 'dummy_annotations_w_freq.csv')
    a = AnnotationTableReader(fname)
    l, b = a.get_annotations('x.wav')
    assert len(l) == 1
    assert l[0] == 0
    assert len(b) == 1
    assert b[0][0] == 1
    assert b[0][1] == 2
    assert b[0][2] == 0.
    assert b[0][3] == 300.

def test_get_annotations_for_file_with_two_annotations():
    fname = os.path.join(path_to_assets, 'dummy_annotations.csv')
    a = AnnotationTableReader(fname)
    l, b = a.get_annotations('y.wav')
    assert len(l) == 2
    assert l[0] == 1
    assert l[1] == 2
    assert len(b) == 2
    assert b[0][0] == 3
    assert b[0][1] == 4
    assert b[0][2] == 0
    import math
    assert b[0][3] == math.inf
    fname = os.path.join(path_to_assets, 'dummy_annotations_w_freq.csv')
    a = AnnotationTableReader(fname)
    l, b = a.get_annotations('y.wav')
    assert len(l) == 2
    assert l[0] == 1
    assert l[1] == 2
    assert len(b) == 2
    assert b[0][0] == 3
    assert b[0][1] == 4
    assert b[0][2] == 4000.
    assert b[0][3] == 6000.

def test_get_annotations_for_file_with_no_annotations():
    fname = os.path.join(path_to_assets, 'dummy_annotations.csv')
    a = AnnotationTableReader(fname)
    l, b = a.get_annotations('z.wav')
    assert len(l) == 0
    assert len(b) == 0

def test_get_maximum_number_of_annotations():
    fname = os.path.join(path_to_assets, 'dummy_annotations.csv')
    a = AnnotationTableReader(fname)
    m = a.get_max_annotations()
    assert m == 2



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'audio_processing' module within the ketos library
"""
import pytest
import os
import numpy as np
import scipy.signal as sg
import ketos.audio_processing.audio_processing as ap
from ketos.audio_processing.audio import AudioSignal
from ketos.audio_processing.spectrogram import Spectrogram, MagSpectrogram

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')



@pytest.mark.test_to_decibel
def test_to_decibel_returns_decibels():
    x = 7
    y = ap.to_decibel(x)
    assert y == 20 * np.log10(x) 

@pytest.mark.test_to_decibel
def test_to_decibel_can_handle_arrays():
    x = np.array([7,8])
    y = ap.to_decibel(x)
    assert np.all(y == 20 * np.log10(x))

@pytest.mark.test_to_decibel
def test_to_decibel_returns_inf_if_input_is_negative():
    x = -7
    y = ap.to_decibel(x)
    assert np.ma.getmask(y) == True

@pytest.mark.test_blur_img
def test_uniform_image_is_unchanged_by_blurring():
    img = np.ones(shape=(10,10), dtype=np.float32)
    img_median = ap.blur_image(img,5,gaussian=False)
    for i in range(img.shape[0]):
        for j in range(img.shape[1]):
            assert img_median[i,j] == img[i,j]
    img_gaussian = ap.blur_image(img,9,gaussian=True)
    np.testing.assert_array_equal(img, img_gaussian)
            
@pytest.mark.test_blur_img
def test_median_filter_can_work_with_kernel_size_greater_than_five():
    img = np.ones(shape=(10,10), dtype=np.float32)
    ap.blur_image(img,13,gaussian=False)

@pytest.mark.test_apply_median_filter
def test_median_filter_works_as_expected():
    img = np.array([[1,1,1],[1,1,1],[1,1,10]], dtype=np.float32)
    img_fil = ap.apply_median_filter(img,row_factor=1,col_factor=1)
    img_res = np.array([[0,0,0],[0,0,0],[0,0,1]], dtype=np.float32)
    np.testing.assert_array_equal(img_fil,img_res)
    img = np.array([[1,1,1],[1,1,1],[1,1,10]], dtype=np.float32)
    img_fil = ap.apply_median_filter(img,row_factor=15,col_factor=1)
    assert img_fil[2,2] == 0
    img = np.array([[1,1,1],[1,1,1],[1,1,10]], dtype=np.float32)
    img_fil = ap.apply_median_filter(img,row_factor=1,col_factor=15)
    assert img_fil[2,2] == 0
    
@pytest.mark.test_apply_preemphasis
def test_preemphasis_has_no_effect_if_coefficient_is_zero():
    sig = np.array([1,2,3,4,5], np.float32)
    sig_new = ap.apply_preemphasis(sig,coeff=0)
    for i in range(len(sig)):
        assert sig[i] == sig_new[i]

def test_make_frames():
    img = np.random.normal(size=100)
    frames1 = ap.make_frames(x=img, winlen=20, winstep=4)
    assert frames1.shape[0] == 21
    assert frames1.shape[1] == 20


def test_prepare_for_binary_cnn():
    n = 1
    l = 2
    specs = list()
    for i in range(n):
        t1 = (i+1) * 2
        t2 = t1 + l + 1
        img = np.ones(shape=(20,30))
        img[t1:t2,:] = 2.5
        s = Spectrogram(image=img)       
        s.annotate(labels=7,boxes=[t1, t2])
        specs.append(s)

    img_wid = 4
    framer = ap.FrameMakerForBinaryCNN(specs=specs, label=7, frame_width=img_wid, step_size=1, signal_width=2)
    x, y, _ = framer.make_frames()
    m = 1 + 20 - 4
    q = 4
    assert y.shape == (m*n,)
    assert x.shape == (m*n, img_wid, specs[0].image.shape[1])
    assert np.all(y[0:q] == 1)
    assert y[4] == 0
    assert np.all(x[0,2,:] == 2.5)
    assert np.all(x[1,1,:] == 2.5)

    framer = ap.FrameMakerForBinaryCNN(specs=specs, label=7, frame_width=img_wid, step_size=1, signal_width=2, equal_rep=True)
    x, y, _ = framer.make_frames()
    assert y.shape == (2*q,)
    assert np.sum(y) == q

@pytest.mark.test_append_specs
def test_append_specs():
    img = np.ones(shape=(20,30))
    s = Spectrogram(image=img)       
    merged = ap.append_specs([s,s,s])
    assert merged.image.shape[0] == 3 * s.image.shape[0]

@pytest.mark.test_filter_isolated_cells
def test_filter_isolated_spots_removes_single_pixels():
    img = np.array([[0,0,1,1,0,0],
                    [0,0,0,1,0,0],
                    [0,1,0,0,0,0],
                    [0,0,0,0,0,0],
                    [0,0,0,1,0,0]])
    
    expected = np.array([[0,0,1,1,0,0],
                        [0,0,0,1,0,0],
                        [0,0,0,0,0,0],
                        [0,0,0,0,0,0],
                        [0,0,0,0,0,0]])
    

    #Struct defines the relationship between a pixel and its neighbors.
    #If a pixel complies with this relationship, it is not removed
    #in this case, if the pixel has any neighbors, it will not be removed.
    struct=np.array([[1,1,1],
                    [1,1,1],
                    [1,1,1]])

    filtered_img = ap.filter_isolated_spots(img,struct)

    assert np.array_equal(filtered_img, expected)

@pytest.mark.test_estimate_audio_signal
def test_estimate_audio_signal(sine_audio):
    n_fft = 20000
    hop = 2000
    winlen = n_fft / sine_audio.rate
    winstep = hop / sine_audio.rate
    spec = MagSpectrogram(audio_signal=sine_audio, winlen=winlen, winstep=winstep)
    img = spec.image
    window = sg.get_window('hamming', n_fft)
    # estimate signal
    audio = ap.estimate_audio_signal(image=img, phase_angle=3.14, n_fft=n_fft, hop=hop, num_iters=25, window=window)
    # check that recovered signal looks like a harmonic with frequency of 2000 Hz
    N = int(6./2000. * sine_audio.rate)
    N = N + N%2 
    f = np.abs(np.fft.rfft(audio[:N]))
    expected = int(2000./(sine_audio.rate/2.)*len(f))
    assert np.argmax(f) == expected


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

import pytest
import datetime
import os
import numpy as np
import scipy.signal as sg
import pandas as pd
import ketos.audio_processing.audio_processing as ap
from ketos.neural_networks.cnn import BasicCNN
from ketos.data_handling.data_handling import to1hot
import ketos.audio_processing.audio as aud
from tensorflow import reset_default_graph

path_to_assets = os.path.join(os.path.dirname(__file__),"assets")


@pytest.fixture
def sine_wave():
    sampling_rate = 44100
    frequency = 2000
    duration = 3
    x = np.arange(duration * sampling_rate)

    signal = 32600*np.sin(2 * np.pi * frequency * x / sampling_rate) 

    return sampling_rate, signal


@pytest.fixture
def square_wave():
    sampling_rate = 44100
    frequency = 2000
    duration = 3
    x = np.arange(duration * sampling_rate)

    signal = 32600 * sg.square(2 * np.pi * frequency * x / sampling_rate) 

    return sampling_rate, signal

@pytest.fixture
def sawtooth_wave():
    sampling_rate = 44100
    frequency = 2000
    duration = 3
    x = np.arange(duration * sampling_rate)

    signal = 32600 * sg.sawtooth(2 * np.pi * frequency * x / sampling_rate) 

    return sampling_rate, signal

@pytest.fixture
def const_wave():
    sampling_rate = 44100
    duration = 3
    x = np.arange(duration * sampling_rate)
    signal = np.ones(len(x))

    return sampling_rate, signal


@pytest.fixture
def sine_wave_file(sine_wave):
    """Create a .wav with the 'sine_wave()' fixture
    
       The file is saved as tests/assets/sine_wave.wav.
       When the tests using this fixture are done, 
       the file is deleted.


       Yields:
            wav_file : str
                A string containing the path to the .wav file.
    """
    wav_file = os.path.join(path_to_assets, "sine_wave.wav")
    rate, sig = sine_wave
    ap.wave.write(wav_file, rate=rate, data=sig)
    
    yield wav_file
    os.remove(wav_file)


@pytest.fixture
def square_wave_file(square_wave):
    """Create a .wav with the 'square_wave()' fixture
    
       The file is saved as tests/assets/square_wave.wav.
       When the tests using this fixture are done, 
       the file is deleted.


       Yields:
            wav_file : str
                A string containing the path to the .wav file.
    """
    wav_file =  os.path.join(path_to_assets, "square_wave.wav")
    rate, sig = square_wave
    ap.wave.write(wav_file, rate=rate, data=sig)

    yield wav_file
    os.remove(wav_file)


@pytest.fixture
def sawtooth_wave_file(sawtooth_wave):
    """Create a .wav with the 'sawtooth_wave()' fixture
    
       The file is saved as tests/assets/sawtooth_wave.wav.
       When the tests using this fixture are done, 
       the file is deleted.


       Yields:
            wav_file : str
                A string containing the path to the .wav file.
    """
    wav_file =  os.path.join(path_to_assets, "sawtooth_wave.wav")
    rate, sig = sawtooth_wave
    ap.wave.write(wav_file, rate=rate, data=sig)

    yield wav_file
    os.remove(wav_file)


@pytest.fixture
def const_wave_file(const_wave):
    """Create a .wav with the 'const_wave()' fixture
    
       The file is saved as tests/assets/const_wave.wav.
       When the tests using this fixture are done, 
       the file is deleted.


       Yields:
            wav_file : str
                A string containing the path to the .wav file.
    """
    wav_file =  os.path.join(path_to_assets, "const_wave.wav")
    rate, sig = const_wave
    ap.wave.write(wav_file, rate=rate, data=sig)

    yield wav_file
    os.remove(wav_file)

@pytest.fixture
def image_2x2():
    image = np.array([[1,2],[3,4]], np.float32)
    return image

@pytest.fixture
def image_3x3():
    image = np.array([[1,2,3],[4,5,6],[7,8,9]], np.float32)
    return image

@pytest.fixture
def image_ones_10x10():
    image = np.ones(shape=(10,10))
    return image

@pytest.fixture
def image_zeros_and_ones_10x10():
    image = np.ones(shape=(10,10))
    for i in range(10):
        for j in range(5):
            image[i,j] = 0
    return image

@pytest.fixture
def datebase_with_one_image_col_and_one_label_col(image_2x2):
    d = {'image': [image_2x2], 'label': [1]}
    df = pd.DataFrame(data=d)
    return df


@pytest.fixture
def datebase_with_one_image_col_and_no_label_col(image_2x2):
    d = {'image': [image_2x2]}
    df = pd.DataFrame(data=d)
    return df


@pytest.fixture
def datebase_with_two_image_cols_and_one_label_col(image_2x2):
    d = {'image1': [image_2x2], 'image2': [image_2x2], 'label': [1]}
    df = pd.DataFrame(data=d)
    return df


@pytest.fixture
def database_prepared_for_NN(image_2x2):
    d = {'image': [image_2x2, image_2x2, image_2x2, image_2x2, image_2x2, image_2x2], 'label': [0,0,0,0,0,0]}
    df = pd.DataFrame(data=d)
    divisions = {"train":(0,3),"validation":(3,4),"test":(4,6)}
    prepared = prepare_database(df, "image", "label", divisions)     
    return prepared

@pytest.fixture
def database_prepared_for_NN_2_classes():
    img1 = np.zeros((20, 20))
    img2 = np.ones((20, 20))
    d = {'image': [img1, img2, img1, img2, img1, img2,
                   img1, img2, img1, img2, img1, img2,
                   img1, img2, img1, img2, img1, img2,
                   img1, img2, img1, img2, img1, img2],
         'label': [0, 1, 0, 1, 0, 1,
                   0, 1, 0, 1, 0, 1,
                   0, 1, 0, 1, 0, 1,
                   0, 1, 0, 1, 0, 1]}
    database = pd.DataFrame(data=d)
    divisions= {"train":(0,12),
                "validation":(12,18),
                "test":(18,len(database))}
    prepared = prepare_database(database=database,x_column="image",y_column="label",
                                divisions=divisions)    
    return prepared


@pytest.fixture
def trained_BasicCNN(database_prepared_for_NN_2_classes):
    d = database_prepared_for_NN_2_classes
    path_to_saved_model = os.path.join(path_to_assets, "saved_models")
    path_to_meta = os.path.join(path_to_saved_model, "trained_BasicCNN")         
    train_x = d["train_x"]
    train_y = d["train_y"]
    validation_x = d["validation_x"]
    validation_y = d["validation_y"]
    test_x = d["test_x"]
    test_y = d["test_y"]
    network = BasicCNN(train_x=train_x, train_y=train_y, validation_x=validation_x, validation_y=validation_y, test_x=test_x, test_y=test_y, batch_size=1, num_labels=2)
    tf_nodes = network.create()
    network.set_tf_nodes(tf_nodes)
    network.train()
    network.save(path_to_meta)
    test_acc = network.accuracy_on_test()
    meta = path_to_meta + ".meta"
    reset_default_graph()
    return meta, path_to_saved_model, test_acc

@pytest.fixture
def sine_audio(sine_wave):
    rate, data = sine_wave
    today = datetime.datetime.today()
    a = aud.TimeStampedAudioSignal(rate=rate, data=data, time_stamp=today, tag="audio")
    return a

@pytest.fixture
def sine_audio_without_time_stamp(sine_wave):
    rate, data = sine_wave
    a = aud.AudioSignal(rate=rate, data=data)
    return a
    
@pytest.fixture
def data_classified_by_nn():
    x = [1, 2, 3, 4, 5, 6] # input data
    x = np.array(x)
    y = [0, 1, 0, 1, 0, 1] # labels
    y = np.array(y)
    w = [[0.8, 0.2], [0.1, 0.9], [0.96, 0.04], [0.49, 0.51], [0.45, 0.55], [0.60, 0.40]] # class weights computed by NN
    w = np.array(w)
    return x,y,w

@pytest.fixture
def data_for_TCN():
    fv0 = np.zeros(64)
    fv1 = np.ones(64)
    x_train = np.array([fv0, fv1, fv0, fv1, fv0, fv1, fv0, fv1, fv0, fv1])
    y_train = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1])
    x_val = np.array([fv0, fv1, fv0, fv1])
    y_val = np.array([0, 1, 0, 1])
    x_test = np.array([fv0, fv1, fv0, fv1])
    y_test = np.array([0, 1, 0, 1])
    return x_train, y_train, x_val, y_val, x_test, y_test
    

def encode_database(database, x_column, y_column):
    image_shape = database[x_column][0].shape
    depth = database[y_column].max() + 1 #number of classes
    database["one_hot_encoding"] = database[y_column].apply(to1hot,depth=depth)
    database["x_flatten"] = database[x_column].apply(lambda x: x.flatten())
    return database, image_shape

def split_database(database, divisions):
    train_data = database[divisions["train"][0]:divisions["train"][1]]
    validation_data = database[divisions["validation"][0]:divisions["validation"][1]]
    test_data = database[divisions["test"][0]:divisions["test"][1]]
    datasets = {"train": train_data,
                "validation": validation_data,
                "test": test_data}
    return datasets

def stack_dataset(dataset, input_shape):
    x = np.vstack(dataset.x_flatten).reshape(dataset.shape[0], input_shape[0], input_shape[1],1).astype(np.float32)
    y = np.vstack(dataset.one_hot_encoding)
    stacked_dataset = {'x': x,
                       'y': y}
    return stacked_dataset

def prepare_database(database, x_column, y_column, divisions):
    encoded_data, input_shape = encode_database(database=database, x_column=x_column, y_column=y_column)
    datasets = split_database(database=encoded_data, divisions=divisions)
    stacked_train = stack_dataset(dataset=datasets["train"], input_shape=input_shape)
    stacked_validation = stack_dataset(dataset=datasets["validation"], input_shape=input_shape)
    stacked_test = stack_dataset(dataset=datasets["test"], input_shape=input_shape)
    stacked_datasets = {"train_x": stacked_train["x"],
                        "train_y": stacked_train["y"],
                        "validation_x": stacked_validation["x"],
                        "validation_y": stacked_validation["y"],
                        "test_x": stacked_test["x"],
                        "test_y": stacked_test["y"]}
    return stacked_datasets



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" EDTCN module within the ketos library

    This module provides utilities to work with Encoder-Decoder Temporal-Convolutional Networks (EDTCN)

    Contents:
        EDTCN class:
"""
import tensorflow as tf
import numpy as np
from ketos.data_handling.data_handling import to1hot, from1hot
from ketos.neural_networks.neural_networks import DataHandler


def channel_normalization(x):
    """ Normalize by the highest activation.

        Args:
            x: tensor
                Tensor containing output values from layer in neural network

        Returns:
            out: tensor
                Values normalized by the highest activation
    """
    max_values = tf.keras.backend.max(tf.keras.backend.abs(x), 2, keepdims=True) + 1e-5
    out = x / max_values
    return out


class EDTCN(DataHandler):
    """ Create an Encoder-Decoder Temporal Convolutional Network (ED-TCN) for classification tasks.

        Implementation adapted from https://github.com/colincsl/TemporalConvolutionalNetworks

        The network architecture can only be constructed if the size of the input data is known.
        Therefore, either num_feat or train_x must be provided.
        In the latter case, the size is automatically determined from the input data.

        Args:
            num_feat: int
                Size of input data
            train_x: numpy array
                Training data
            train_y: numpy array
                Labels for training data. Can be integers or 1-hot encoded
            validation_x: numpy array
                Validation data
            validation_y: numpy array
                Labels for validation data. Can be integers or 1-hot encoded
            test_x: numpy array
                Test data
            test_y: numpy array
                Labels for test data. Can be integers or 1-hot encoded
            num_labels: int
                Number of labels
            max_len: int
                The input data will be split into chuncks of size max_len. 
                Thus, max_len effectively limits the extent of the memory of the network. 
            batch_size: int
                The number of examples in each batch
            num_epochs: int
                The number of epochs
            keep_prob: float
                Probability of keeping weights during training. Set keep_prob to 1.0 to disable drop-out (default).
            verbosity: int
                Verbosity level (0: no messages, 1: warnings only, 2: warnings and diagnostics)

        Attributes:
            model: tf.keras.models.Model
                Neural network model
            class_weights_func: tf.keras.backend.function
                Function to compute classification weights

        Example:

            >>> # initialize EDTCN for classifying feature vectors of size 64
            >>> from ketos.neural_networks.edtcn import EDTCN
            >>> tcn = EDTCN(num_feat=64)
            >>> print(tcn.num_feat)
            64
    """
    def __init__(self, num_feat=None, train_x=None, train_y=None, validation_x=None, validation_y=None,
                 test_x=None, test_y=None, num_labels=2, max_len=500, batch_size=4, 
                 num_epochs=100, keep_prob=0.7, verbosity=0):

        assert (num_feat is not None) or (train_x is not None), "num_feat or train_x must be provided"

        self.batch_size = batch_size
        self.num_epochs = num_epochs
        self.keep_prob = keep_prob
        self.verbosity = verbosity
        self.max_len = max_len

        self.model = None
        self.class_weights_func = None

        self.max_len = max_len

        super(EDTCN, self).__init__(train_x=train_x, train_y=train_y, 
                validation_x=validation_x, validation_y=validation_y,
                test_x=test_x, test_y=test_y, num_labels=num_labels)

        if num_feat is None:
            x, _ = self.get_training_data()
            x = np.squeeze(x)
            self.num_feat = x.shape[-1]
        else:
            self.num_feat = num_feat

    def set_verbosity(self, verbosity):
        """Set verbosity level.

            0: no messages
            1: warnings only
            2: warnings and diagnostics

            Args:
                verbosity: int
                    Verbosity level.        
        """
        self.verbosity = verbosity

    def create(self, n_nodes=[16, 32], conv_len=16, max_len=None):
        """Create the Neural Network structure.

            Args:
                n_nodes: tuple(int)
                    Number of output filters in each 1D convolutional layer.
                    (The number of convolutional layers is given by the length of n_nodes)
                conv_len: int
                    Length of the 1D convolution window.
                max_len: int
                    The input data will be split into chuncks of size max_len. 
                    Thus, max_len effectively limits the extent of the memory of the network. 

            Example:

                >>> # initialize EDTCN for classifying feature vectors of size 64
                >>> from ketos.neural_networks.edtcn import EDTCN
                >>> tcn = EDTCN(num_feat=64)
                >>> # create network with default architecture
                >>> tcn.create()
        """
        dropout_rate = 1.0 - self.keep_prob

        if max_len is not None:
            self.max_len = max_len

        # ensure that max_len is divisible by four (as this seems to be required by keras.layers.Input)
        if self.max_len % 4 > 0:
            self.max_len += 4 - self.max_len % 4
            if self.verbosity >= 1: 
                print(' Warning: max_len must be divisible by 4; max_len has been adjust to {0}'.format(self.max_len))

        n_feat = self.num_feat 
        n_layers = len(n_nodes)
        n_classes = self.num_labels
        
        # --- Input layer ---
        inputs = tf.keras.layers.Input(shape=(self.max_len, n_feat))
        model = inputs

        # ---- Encoder ----
        for i in range(n_layers):
            model = tf.keras.layers.Convolution1D(n_nodes[i], conv_len, padding='same')(model) # 1D convolutional layer
            model = tf.keras.layers.SpatialDropout1D(rate=dropout_rate)(model) # Spatial 1D version of Dropout
            model = tf.keras.layers.Activation('relu')(model) # apply activation
            model = tf.keras.layers.Lambda(channel_normalization, name="encoder_norm_{}".format(i))(model)
            model = tf.keras.layers.MaxPooling1D(2)(model) # max pooling

        # ---- Decoder ----
        for i in range(n_layers):
            model = tf.keras.layers.UpSampling1D(2)(model)
            model = tf.keras.layers.Convolution1D(n_nodes[-i-1], conv_len, padding='same')(model)
            model = tf.keras.layers.SpatialDropout1D(rate=dropout_rate)(model)
            model = tf.keras.layers.Activation('relu')(model)
            model = tf.keras.layers.Lambda(channel_normalization, name="decoder_norm_{}".format(i))(model)

        # Output fully connected layer
        model = tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(units=n_classes, activation="softmax" ))(model)

        # create and compile model
        model = tf.keras.models.Model(inputs=inputs, outputs=model)
        model.compile(loss='categorical_crossentropy', optimizer="rmsprop", sample_weight_mode="temporal", metrics=['accuracy'])

        self.model = model

        # make function to retrieve classification weights
        inp = self.model.input                
        output = self.model.layers[-1].output
        self.class_weights_func = tf.keras.backend.function([inp, tf.keras.backend.learning_phase()], [output])

    def train(self, batch_size=None, num_epochs=None, equal_weighting=False):
        """ Train the neural network on the training set.

            Divide the training set in batches of size batch_size. 

            Args:
                batch_size: int
                    Batch size. Overwrites batch size specified at initialization.
                num_epochs: int
                    Number of epochs: Overwrites number of epochs specified at initialization.
                equal_weighting: bool
                    Correct for any imbalance in the occurrence of class examples in the training data, so 
                    that all classes are weighted equally in calculation of the loss function. 
                    Obs: only implemeted for binary classification! 

            Returns:
                history: 
                    Keras training history.
        """
        if batch_size is None:
            batch_size = self.batch_size
        if num_epochs is None:
            num_epochs = self.num_epochs

        x, y = self.get_training_data()
        x_val, y_val = self.get_validation_data()

        # ensure equal weighting of 0s and 1s
        if equal_weighting:

            y1 = from1hot(y)
            assert np.max(y) == 1, 'Equal weighting has only been implemented for binary classification'

            n0 = np.sum(y1==0)
            n = y1.shape[0]
            w0 = float(n - n0) / float(n)
            sample_weight = y1 * (1. - w0) - (y1 - 1) * w0
            sample_weight = self._reshape(sample_weight[:, np.newaxis])
            sample_weight = np.squeeze(sample_weight)            
        
        else:
            sample_weight = None

        if x_val is not None and y_val is not None:
            x_val = self._reshape(x_val)
            y_val = self._reshape(y_val)
            val_data = (x_val, y_val)
        else:
            val_data = None

        x = self._reshape(x)
        y = self._reshape(y)

        history = self.model.fit(x=x, y=y, batch_size=batch_size, epochs=num_epochs,\
                    verbose=self.verbosity, validation_data=val_data, sample_weight=sample_weight)   
        return history     

    def get_predictions(self, x):
        """ Predict labels by running the model on x

            Args:
                x: tensor
                    Tensor containing the input data.
                
            Returns:
                results: vector
                    A vector containing the predicted labels.                
        """
        x = np.array(x)

        if np.ndim(x) == 1:
            x = x[np.newaxis, :]

        orig_len = x.shape[0]

        x = self._reshape(x)
        results = self.model.predict(x=x)
        results = np.reshape(results, newshape=(results.shape[0]*results.shape[1], results.shape[2]))

        results = from1hot(results)
        results = results[:orig_len]

        if results.shape[0] == 1:
            results = results[0]

        return results

    def get_class_weights(self, x):
        """ Compute classification weights by running the model on x.

            Args:
                x: tensor
                    Tensor containing the input data.
                
            Returns:
                results: vector
                    A vector containing the classification weights. 
        """
        x = np.array(x)

        if np.ndim(x) == 1:
            x = x[np.newaxis, :]

        orig_len = x.shape[0]

        x = self._reshape(x)        
        w = np.array(self.class_weights_func([x, False]))
        w = np.squeeze(w)
        if np.ndim(w) == 2:
            w = w[np.newaxis,:,:]
            
        w = np.reshape(w, newshape=(w.shape[0]*w.shape[1], w.shape[2]))

        w = w[:orig_len]
        if w.shape[0] == 1:
            w = w[0]

        return w

    def _reshape(self, a):
        """ Split the data into chunks with size max_len.
            
            Args:
                a: numpy array
                    Array containing the data to be split.
        """
        if np.ndim(a) > 2:
            a = np.squeeze(a)

        n = self.max_len
        orig_len = a.shape[0]
        nsegs = int(np.ceil(orig_len / n))
        new_len = nsegs * n

        pad_shape = np.array([new_len - orig_len], dtype=np.int32)
        pad_shape = np.append(pad_shape, a.shape[1:])
        a = np.append(a, np.zeros(shape=pad_shape), axis=0)

        new_shape = np.array([nsegs, n], dtype=np.int32)
        new_shape = np.append(new_shape, a.shape[1:])
        a = np.reshape(a=a[:new_len], newshape=new_shape)

        return a

    def save(self, destination):
        """ Save the model to destination

            Args:
                destination: str
                    Path to the file in which the model will be saved. 

            Returns:
                None.
        
        """
        tf.keras.models.save_model(model=self.model, filepath=destination)

    # TODO: The current implementation of the load() method does not work.
    #       It gives the error message "in channel_normalization max_values 
    #       = tf.keras.backend.max(tf.keras.backend.abs(x), 2, keepdims=True) 
    #       + 1e-5 NameError: name 'tf' is not defined"

    def load(self, path):
        """Load the Neural Network structure and weights from a saved model.

            See the save() method. 

            Args:
                path: str
                    Path to the saved model.
        """
        self.model = tf.keras.models.load_model(filepath=path)

    def save_weights(self, path):
        """ Save the model weights to destination

            Args:
                destination: str
                    Path to the file in which the weights will be saved. 

            Returns:
                None.
        """
        self.model.save_weights(filepath=path)        

    def load_weights(self, path):
        """ Load the model weights

            Args:
                destination: str
                    Path to the file in which the weights are stored.

            Returns:
                None.
        """
        self.model.load_weights(filepath=path)        


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" audio module within the ketos library

    This module provides utilities to work with audio data.

    Contents:
        AudioSignal class: 
        TimeStampedAudioSignal class
"""

import os
import numpy as np
import datetime
import math
import scipy.io.wavfile as wave
from scipy import interpolate
from ketos.utils import morlet_func
import ketos.audio_processing.audio_processing as ap
import matplotlib.pyplot as plt
from scipy.integrate import quadrature
from scipy.stats import norm
from tqdm import tqdm
from ketos.audio_processing.annotation import AnnotationHandler
from ketos.data_handling.data_handling import read_wave
from ketos.utils import ensure_dir


class AudioSignal(AnnotationHandler):
    """ Audio signal

        Args:
            rate: float
                Sampling rate in Hz
            data: 1d numpy array
                Audio data 
            tag: str
                Meta-data string (optional)
            tstart: float
                Start time in seconds (optional)

        Attributes:
            rate: float
                Sampling rate in Hz
            data: 1d numpy array
                Audio data 
            tag: str
                Meta-data string
            tmin: float
                Start time in seconds              
            file_dict: dict
                Wave files used to generate this spectrogram
            file_vector: 1d numpy array
                Associates a particular wave file with each time bin in the spectrogram
            time_vector: 1d numpy array
                Associated a particular time within a wave file with each time bin in the spectrogram                
    """
    def __init__(self, rate, data, tag='', tstart=0):

        self.rate = float(rate)
        self.data = data.astype(dtype=np.float32)
        self.tag = tag
        self.tmin = tstart

        super(AudioSignal, self).__init__() # initialize AnnotationHandler

        n = self.data.shape[0]
        self.time_vector = (1. / self.rate) * np.arange(n) + self.tmin
        self.file_vector = np.zeros(n)
        self.file_dict = {0: tag}

    @classmethod
    def from_wav(cls, path, channel=0):
        """ Generate audio signal from wave file

            The tag attribute will be set equal to the file name.

            Args:
                path: str
                    Path to input wave file
                channel: int
                    In the case of stereo recordings, this argument is used 
                    to specify which channel to read from. Default is 0.

            Returns:
                Instance of AudioSignal
                    Audio signal from wave file

            Example:

            >>> from ketos.audio_processing.audio import AudioSignal
            >>> # read audio signal from wav file
            >>> a = AudioSignal.from_wav('ketos/tests/assets/grunt1.wav')
            >>> # show signal
            >>> fig = a.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/audio_grunt1.png")

            .. image:: ../../../../ketos/tests/assets/tmp/audio_grunt1.png

        """        
        rate, data = read_wave(file=path, channel=channel)
        _, fname = os.path.split(path)
        return cls(rate, data, fname)

    @classmethod
    def gaussian_noise(cls, rate, sigma, samples, tag=''):
        """ Generate Gaussian noise signal

            Args:
                rate: float
                    Sampling rate in Hz
                sigma: float
                    Standard deviation of the signal amplitude
                samples: int
                    Length of the audio signal given as the number of samples
                tag: str
                    Meta-data string (optional)

            Returns:
                Instance of AudioSignal
                    Audio signal sampling of Gaussian noise

            Example:

            >>> from ketos.audio_processing.audio import AudioSignal
            >>> # create gaussian noise with sampling rate of 10 Hz, standard deviation of 2.0 and 1000 samples
            >>> a = AudioSignal.gaussian_noise(rate=10, sigma=2.0, samples=1000)
            >>> # show signal
            >>> fig = a.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/audio_noise.png")

            .. image:: ../../../../ketos/tests/assets/tmp/audio_noise.png

        """        
        assert sigma > 0, "sigma must be strictly positive"

        if tag == '':
            tag = "Gaussian_noise_sigma{0:.3f}".format(sigma)

        y = np.random.normal(loc=0, scale=sigma, size=samples)
        return cls(rate=rate, data=y, tag=tag)

    @classmethod
    def morlet(cls, rate, frequency, width, samples=None, height=1, displacement=0, dfdt=0, tag=''):
        """ Audio signal with the shape of the Morlet wavelet

            Uses :func:`util.morlet_func` to compute the Morlet wavelet.

            Args:
                rate: float
                    Sampling rate in Hz
                frequency: float
                    Frequency of the Morlet wavelet in Hz
                width: float
                    Width of the Morlet wavelet in seconds (sigma of the Gaussian envelope)
                samples: int
                    Length of the audio signal given as the number of samples (if no value is given, samples = 6 * width * rate)
                height: float
                    Peak value of the audio signal
                displacement: float
                    Peak position in seconds
                dfdt: float
                    Rate of change in frequency as a function of time in Hz per second.
                    If dfdt is non-zero, the frequency is computed as 
                        
                        f = frequency + (time - displacement) * dfdt 

                tag: str
                    Meta-data string (optional)

            Returns:
                Instance of AudioSignal
                    Audio signal sampling of the Morlet wavelet 

            Examples:

            >>> from ketos.audio_processing.audio import AudioSignal
            >>> # create a Morlet wavelet with frequency of 3 Hz and 1-sigma width of envelope set to 2.0 seconds
            >>> wavelet1 = AudioSignal.morlet(rate=100., frequency=3., width=2.0)
            >>> # show signal
            >>> fig = wavelet1.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/morlet_standard.png")

            .. image:: ../../../../ketos/tests/assets/tmp/morlet_standard.png

            >>> # create another wavelet, but with frequency increasing linearly with time
            >>> wavelet2 = AudioSignal.morlet(rate=100., frequency=3., width=2.0, dfdt=0.3)
            >>> # show signal
            >>> fig = wavelet2.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/morlet_dfdt.png")

            .. image:: ../../../../ketos/tests/assets/tmp/morlet_dfdt.png

        """        
        if samples is None:
            samples = int(6 * width * rate)

        N = int(samples)

        # compute Morlet function at N equally spaced points
        dt = 1. / rate
        stop = (N-1.)/2. * dt
        start = -stop
        time = np.linspace(start, stop, N)
        y = morlet_func(time=time, frequency=frequency, width=width, displacement=displacement, norm=False, dfdt=dfdt)        
        y *= height
        
        if tag == '':
            tag = "Morlet_f{0:.0f}Hz_s{1:.3f}s".format(frequency, width) # this is just a string with some helpful info

        return cls(rate=rate, data=np.array(y), tag=tag)

    @classmethod
    def cosine(cls, rate, frequency, duration=1, height=1, displacement=0, tag=''):
        """ Audio signal with the shape of a cosine function

            Args:
                rate: float
                    Sampling rate in Hz
                frequency: float
                    Frequency of the Morlet wavelet in Hz
                duration: float
                    Duration of the signal in seconds
                height: float
                    Peak value of the audio signal
                displacement: float
                    Phase offset in fractions of 2*pi
                tag: str
                    Meta-data string (optional)

            Returns:
                Instance of AudioSignal
                    Audio signal sampling of the cosine function 

            Examples:

            >>> from ketos.audio_processing.audio import AudioSignal
            >>> # create a Cosine wave with frequency of 7 Hz
            >>> cos = AudioSignal.cosine(rate=1000., frequency=7.)
            >>> # show signal
            >>> fig = cos.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/cosine_audio.png")

            .. image:: ../../../../ketos/tests/assets/tmp/cosine_audio.png

        """        
        N = int(duration * rate)

        # compute cosine function at N equally spaced points
        dt = 1. / rate
        stop = (N-1.)/2. * dt
        start = -stop
        time = np.linspace(start, stop, N)
        x = (time * frequency + displacement) * 2 * np.pi
        y = height * np.cos(x)
        
        if tag == '':
            tag = "cosine_f{0:.0f}Hz".format(frequency) # this is just a string with some helpful info

        return cls(rate=rate, data=np.array(y), tag=tag)

    def get_data(self):
        """ Get the underlying data contained in this object.
            
            Returns:
                self.data: numpy array
                    Data
        """
        return self.data

    def get_time_vector(self):
        """ Get the audio signal's time vector

            Returns:
                v: 1d numpy array
                    Time vector
        """
        v = self.time_vector
        return v

    def get_file_vector(self):
        """ Get the audio signal's file vector

            Returns:
                v: 1d numpy array
                    File vector
        """
        v = self.file_vector
        return v

    def get_file_dict(self):
        """ Get the audio signal's file dictionary

            Returns:
                d: dict
                    File dictionary
        """
        d = self.file_dict
        return d

    def make_frames(self, winlen, winstep, zero_padding=False, even_winlen=False):
        """ Split the signal into frames of length 'winlen' with consecutive 
            frames being shifted by an amount 'winstep'. 
            
            If 'winstep' < 'winlen', the frames overlap.

        Args: 
            signal: AudioSignal
                The signal to be framed.
            winlen: float
                The window length in seconds.
            winstep: float
                The window step (or stride) in seconds.
            zero_padding: bool
                If necessary, pad the signal with zeros at the end to make sure that all frames have equal number of samples.
                This assures that sample are not truncated from the original signal.

        Returns:
            frames: numpy array
                2-d array with padded frames.
        """
        rate = self.rate
        sig = self.data

        winlen = int(round(winlen * rate))
        winstep = int(round(winstep * rate))

        if even_winlen and winlen%2 != 0:
            winlen += 1

        frames = ap.make_frames(sig, winlen, winstep, zero_padding)
        return frames

    def to_wav(self, path, auto_loudness=True):
        """ Save audio signal to wave file

            Args:
                path: str
                    Path to output wave file
                auto_loudness: bool
                    Automatically amplify the signal so that the 
                    maximum amplitude matches the full range of 
                    a 16-bit wav file (32760)
        """        
        ensure_dir(path)
        
        if auto_loudness:
            m = max(1, np.max(np.abs(self.data)))
            s = 32760 / m
        else:
            s = 1

        wave.write(filename=path, rate=int(self.rate), data=(s*self.data).astype(dtype=np.int16))

    def empty(self):
        """ Check if the signal contains any data

            Returns:
                bool
                    True if the length of the data array is zero or array is None
        """  
        if self.data is None:
            return True
        elif len(self.data) == 0:
            return True
        
        return False

    def duration(self):
        """ Signal duration in seconds

            Returns:
                s: float
                   Signal duration in seconds
        """    
        s = float(len(self.data)) / float(self.rate)
        return s

    def max(self):
        """ Maximum value of the signal

            Returns:
                v: float
                   Maximum value of the data array
        """    
        v = max(self.data)
        return v

    def min(self):
        """ Minimum value of the signal

            Returns:
                v: float
                   Minimum value of the data array
        """    
        v = min(self.data)
        return v

    def std(self):
        """ Standard deviation of the signal

            Returns:
                v: float
                   Standard deviation of the data array
        """   
        v = np.std(self.data) 
        return v

    def average(self):
        """ Average value of the signal

            Returns:
                v: float
                   Average value of the data array
        """   
        v = np.average(self.data)
        return v

    def median(self):
        """ Median value of the signal

            Returns:
                v: float
                   Median value of the data array
        """   
        v = np.median(self.data)
        return v

    def plot(self):
        """ Plot the signal with proper axes ranges and labels
            
            Example:
            
            >>> from ketos.audio_processing.audio import AudioSignal
            >>> # create a morlet wavelet
            >>> a = AudioSignal.morlet(rate=100, frequency=5, width=1)
            >>> # plot the wave form
            >>> fig = a.plot()

            .. image:: ../../_static/morlet.png

        """
        fig, ax = plt.subplots(nrows=1)
        start = 0.5 / self.rate
        stop = self.duration() - 0.5 / self.rate
        num = len(self.data)
        ax.plot(np.linspace(start=start, stop=stop, num=num), self.data)
        ax = plt.gca()
        ax.set_xlabel('Time (s)')
        ax.set_ylabel('Signal')
        return fig

    def _selection(self, begin, end):
        """ Convert time range to sample range.

            Args:
                begin: float
                    Start time of selection window in seconds
                end: float
                    End time of selection window in seconds

            Returns:
                i1: int
                    Start sample no.
                i2: int
                    End sample no.
        """   
        i1 = 0
        i2 = len(self.data)

        if begin is not None:
            begin = max(0, begin)
            i1 = int(begin * self.rate)
            i1 = max(i1, 0)

        if end is not None:
            end = min(self.duration(), end)
            i2 = int(end * self.rate)
            i2 = min(i2, len(self.data))

        return i1, i2

    def _crop(self, i1, i2):
        """ Select a portion of the audio data

            Args:
                i1: int
                    Start bin of selection window
                end: int
                    End bin of selection window

            Returns:
                cropped_data: numpy array
                   Selected portion of the audio data
        """
        if i2 > i1:
            self.data = self.data[i1:i2] 
        else:
            self.data = None           

    def crop(self, begin=None, end=None, make_copy=False):
        """ Clip audio signal

            Args:
                begin: float
                    Start time of selection window in seconds
                end: float
                    End time of selection window in seconds
                make_copy: bool
                    Create a new instance instead of modifying the present instance

            Returns:
                cropped_signal: AudioSignal
                    None, unless make_copy is True

            Example:
            
            >>> from ketos.audio_processing.audio import AudioSignal
            >>> # create a morlet wavelet
            >>> a1 = AudioSignal.morlet(rate=100, frequency=5, width=1)
            >>> # crop the first 1 second
            >>> a2 = a1.crop(end=1.0, make_copy=True)
            >>> # show the wave forms
            >>> fig = a1.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/audio_orig.png")
            >>> fig = a2.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/audio_cropped.png")

            .. image:: ../../../../ketos/tests/assets/tmp/audio_orig.png

            .. image:: ../../../../ketos/tests/assets/tmp/audio_cropped.png

        """ 
        if make_copy:
            x = self.copy()
        else:
            x = self

        i1, i2 = x._selection(begin, end)
        x._crop(i1, i2)

        if make_copy:
            return x

    def clip(self, boxes):
        """ Extract boxed intervals from audio signal.

            After clipping, this instance contains the remaining part of the audio signal.

            Args:
                boxes: numpy array
                    2d numpy array with shape (?,2)   

            Returns:
                segs: list(AudioSignal)
                    List of clipped audio signals.                

            Example:
            
            >>> from ketos.audio_processing.audio import AudioSignal
            >>> # create a morlet wavelet
            >>> a = AudioSignal.morlet(rate=100, frequency=5, width=1)
            >>> # clip segment between 1.0-1.8 sec and again between 3.1-3.8 sec
            >>> segs = a.clip(boxes=[[1.0, 1.8],[3.1, 3.8]])
            >>> # show the wave forms
            >>> fig = a.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/audio_whats_left.png")
            >>> fig = segs[0].plot()
            >>> fig.savefig("ketos/tests/assets/tmp/audio_clip_1.png")
            >>> fig = segs[1].plot()
            >>> fig.savefig("ketos/tests/assets/tmp/audio_clip_2.png")

            .. image:: ../../../../ketos/tests/assets/tmp/audio_whats_left.png

            .. image:: ../../../../ketos/tests/assets/tmp/audio_clip_1.png

            .. image:: ../../../../ketos/tests/assets/tmp/audio_clip_2.png

        """ 

        if np.ndim(boxes) == 1:
            boxes = [boxes]

        # sort boxes in chronological order
        sorted(boxes, key=lambda box: box[0])

        boxes = np.array(boxes)
        N = boxes.shape[0]

        # get cuts
        segs = list()

        # loop over boxes
        t1, t2 = list(), list()
        for i in range(N):
            
            b = boxes[i]

            assert len(b) >= 2, "box must have dimension 2 or greater"

            begin = b[0]
            end = b[1]
            t1i, t2i = self._selection(begin, end)

            data = self.data[t1i:t2i]
            seg = AudioSignal(rate=self.rate, data=data)

            segs.append(seg)
            t1.append(t1i)
            t2.append(t2i)

        # complement
        t2 = np.insert(t2, 0, 0)
        t1 = np.append(t1, len(self.data))
        t2max = 0
        for i in range(len(t1)):
            t2max = max(t2[i], t2max)
            if t2max < t1[i]:
                if t2max == 0:
                    data_c = self.data[t2max:t1[i]]
                else:
                    data_c = np.append(data_c, self.data[t2max:t1[i]], axis=0)

        self.data = data_c
        self.tmin = 0

        return segs

    def segment(self, length, pad=False, keep_time=False):
        """ Split the audio signal into a number of equally long segments.

            Args:
                length: float
                    Duration of each segment in seconds
                pad: bool
                    If True, pad spectrogram with zeros if necessary to ensure 
                    that bins are used.
                keep_time: bool
                    If True, the extracted segments keep the time from the present instance. 
                    If False, the time axis of each extracted segment starts at t=0

            Returns:
                segs: list
                    List of segments
        """
        if length >= self.duration():
            return [self]

        # split data array into segments
        frames = self.make_frames(winlen=length, winstep=length, zero_padding=pad)

        # create audio signals
        segs = list()
        tstart = self.tmin
        for f in frames:

            if not keep_time:
                tstart = 0

            # audio signal
            a = AudioSignal(rate=self.rate, data=f, tag=self.tag, tstart=tstart)

            # handle annotations
            for l,b in zip(self.labels, self.boxes):
                if b[0] < a.tmin + a.duration() and b[1] > a.tmin:            
                    a.annotate(labels=l, boxes=b)
    
            segs.append(a)
            tstart += length 

        return segs

    def append(self, signal, delay=None, n_smooth=0, max_length=None):
        """ Append another audio signal to this signal.

            The two audio signals must have the same samling rate.
            
            If delay is None or 0, a smooth transition is made between the 
            two signals. The width of the smoothing region where the two signals 
            overlap is specified via the argument n_smooth.

            Note that the current implementation of the smoothing procedure is 
            quite slow, so it is advisable to use small overlap regions.

            If n_smooth is 0, the two signals are joint without any smoothing.

            If delay > 0, a signal with zero sound intensity and duration 
            delay is added between the two audio signals. 

            If the length of the combined signal exceeds max_length, only the 
            first max_length samples will be kept. 

            Args:
                signal: AudioSignal
                    Audio signal to be merged.
                delay: float
                    Delay between the two audio signals in seconds. 
                n_smooth: int
                    Width of the smoothing/overlap region (number of samples).
                max_length: int
                    Maximum length of the combined signal (number of samples).

            Returns:
                append_time: float
                    Start time of appended part in seconds from the beginning of the original signal.

            Example:

            >>> from ketos.audio_processing.audio import AudioSignal
            >>> # create a morlet wavelet
            >>> mor = AudioSignal.morlet(rate=100, frequency=5, width=1)
            >>> # create a cosine wave
            >>> cos = AudioSignal.cosine(rate=100, frequency=3, duration=4)
            >>> # append the cosine wave to the morlet wavelet, using a overlap of 100 bins
            >>> mor.append(signal=cos, n_smooth=100)
            5.0
            >>> # show the wave form
            >>> fig = mor.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/morlet_cosine.png")

            .. image:: ../../../../ketos/tests/assets/tmp/morlet_cosine.png

        """   
        assert self.rate == signal.rate, "Cannot merge audio signals with different sampling rates."

        # if appending signal to itself, make a copy
        if signal is self:
            signal = self.copy()

        if delay is None:
            delay = 0

        delay = max(0, delay)
        if max_length is not None:
            delay = min(max_length / self.rate, delay)

        # ensure that overlap region is shorter than either signals
        n_smooth = min(n_smooth, len(self.data) - 1)
        n_smooth = min(n_smooth, len(signal.data) - 1)

        # compute total length
        len_tot = self.merged_length(signal, delay, n_smooth)

        append_time = len(self.data) / self.rate

        # extract data from overlap region
        if delay == 0 and n_smooth > 0:

            # signal 1
            a = self.split(-n_smooth)

            # signal 2
            b = signal.split(n_smooth)

            # superimpose a and b
            # TODO: If possible, vectorize this loop for faster execution
            # TODO: Cache values returned by _smoothclamp to avoid repeated calculation
            # TODO: Use coarser binning for smoothing function to speed things up even more
            c = np.empty(n_smooth)
            for i in range(n_smooth):
                w = _smoothclamp(i, 0, n_smooth-1)
                c[i] = (1.-w) * a.data[i] + w * b.data[i]
            
            append_time = len(self.data) / self.rate

            # append
            self.data = np.append(self.data, c)

        elif delay > 0:
            z = np.zeros(int(delay * self.rate))
            self.data = np.append(self.data, z)
            append_time = len(self.data) / self.rate
            
        self.data = np.append(self.data, signal.data) 
        
        assert len(self.data) == len_tot # check that length of merged signal is as expected

        # mask inserted zeros
        if delay > 0:
            self.data = np.ma.masked_values(self.data, 0)

        # remove all appended data from signal        
        if max_length is not None:
            if len_tot > max_length:
                self._crop(i1=0, i2=max_length)
                i2 = len(signal.data)
                i1 = max(0, i2 - (len_tot - max_length))
                signal._crop(i1=i1, i2=i2)
            else:
                signal.data = None
        else:
            signal.data = None
        
        return append_time

    def split(self, s):
        """ Split audio signal.

            After splitting, this instance contains the remaining part of the audio signal.        

            Args:
                s: int
                    If s >= 0, select samples [:s]. If s < 0, select samples [-s:]
                    
            Returns:
                Instance of AudioSignal
                    Selected part of the audio signal.
        """   
        if s is None or s == math.inf:
            s = len(self.data)
        
        if s >= 0:
            v = self.data[:s]
            self._crop(i1=s,i2=len(self.data))
        else:
            v = self.data[s:]
            self._crop(i1=0,i2=len(self.data)+s)
 
        return AudioSignal(rate=self.rate, data=v, tag=self.tag)

    def merged_length(self, signal=None, delay=None, n_smooth=None):
        """ Compute sample size of merged signal (without actually merging the signals)

            Args:
                signal: AudioSignal
                    Audio signal to be merged
                delay: float
                    Delay between the two audio signals in seconds.
        """   
        if signal is None:
            return len(self.data)

        assert self.rate == signal.rate, "Cannot merge audio signals with different sampling rates."

        if delay is None:
            delay = 0
            
        if n_smooth is None:
            n_smooth = 0

        m = len(self.data)
        n = len(signal.data)
        l = m + n
        
        if delay > 0:
            l += int(delay * self.rate)
        elif delay == 0:
            l -= n_smooth
        
        return l

    def add_gaussian_noise(self, sigma):
        """ Add Gaussian noise to the signal

            Args:
                sigma: float
                    Standard deviation of the gaussian noise

            Example:

            >>> from ketos.audio_processing.audio import AudioSignal
            >>> # create a morlet wavelet
            >>> morlet = AudioSignal.morlet(rate=100, frequency=2.5, width=1)
            >>> morlet_pure = morlet.copy() # make a copy
            >>> # add some noise
            >>> morlet.add_gaussian_noise(sigma=0.3)
            >>> # show the wave form
            >>> fig = morlet_pure.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/morlet_wo_noise.png")
            >>> fig = morlet.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/morlet_w_noise.png")

            .. image:: ../../../../ketos/tests/assets/tmp/morlet_wo_noise.png

            .. image:: ../../../../ketos/tests/assets/tmp/morlet_w_noise.png

        """
        noise = AudioSignal.gaussian_noise(rate=self.rate, sigma=sigma, samples=len(self.data))
        self.add(noise)

    def add(self, signal, delay=0, scale=1):
        """ Add the amplitudes of the two audio signals.
        
            The audio signals must have the same sampling rates.

            The summed signal always has the same length as the present instance.

            If the audio signals have different lengths and/or a non-zero delay is selected, 
            only the overlap region will be affected by the operation.
            
            If the overlap region is empty, the original signal is unchanged.

            Args:
                signal: AudioSignal
                    Audio signal to be added
                delay: float
                    Shift the audio signal by this many seconds
                scale: float
                    Scaling factor for signal to be added

            Example:

            >>> from ketos.audio_processing.audio import AudioSignal
            >>> # create a morlet wavelet
            >>> mor = AudioSignal.morlet(rate=100, frequency=5, width=1)
            >>> # create a cosine wave
            >>> cos = AudioSignal.cosine(rate=100, frequency=3, duration=4)
            >>> # add the cosine on top of the morlet wavelet, with a delay of 2 sec and a scaling factor of 0.3
            >>> mor.add(signal=cos, delay=2.0, scale=0.3)
            >>> # show the wave form
            >>> fig = mor.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/morlet_cosine_added.png")

            .. image:: ../../../../ketos/tests/assets/tmp/morlet_cosine_added.png

        """
        assert self.rate == signal.rate, "Cannot add audio signals with different sampling rates."

        if delay >= 0:
            i_min = int(delay * self.rate)
            j_min = 0
            i_max = min(self.data.shape[0], signal.data.shape[0] + i_min)
            j_max = min(signal.data.shape[0], self.data.shape[0] - i_min)
            
        else:
            i_min = 0
            j_min = int(-delay * self.rate)
            i_max = min(self.data.shape[0], signal.data.shape[0] - j_min)
            j_max = min(signal.data.shape[0], self.data.shape[0] + j_min)
            
        if i_max > i_min and i_max > 0 and j_max > j_min and j_max > 0:
            self.data[i_min:i_max] += scale * signal.data[j_min:j_max]

    def resample(self, new_rate):
        """ Resample the acoustic signal with an arbitrary sampling rate.

        Note: Code adapted from Kahl et al. (2017)
              Paper: http://ceur-ws.org/Vol-1866/paper_143.pdf
              Code:  https://github.com/kahst/BirdCLEF2017/blob/master/birdCLEF_spec.py  

        Args:
            new_rate: int
                New sampling rate in Hz
        """
        if len(self.data) < 2:
            self.rate = new_rate

        else:                
            orig_rate = self.rate
            sig = self.data

            duration = sig.shape[0] / orig_rate

            time_old  = np.linspace(0, duration, sig.shape[0])
            time_new  = np.linspace(0, duration, int(sig.shape[0] * new_rate / orig_rate))

            interpolator = interpolate.interp1d(time_old, sig.T)
            new_audio = interpolator(time_new).T

            new_sig = np.round(new_audio).astype(sig.dtype)

            self.rate = new_rate
            self.data = new_sig

    def copy(self):
        """ Makes a hard copy of the present instance.

            Returns:
                Instance of TimeStampedAudioSignal
                    Copied signal
        """                
        data = np.copy(self.data)
        return AudioSignal(rate=self.rate, data=data, tag=self.tag, tstart=self.tmin)

def _smoothclamp(x, mi, mx): 
        """ Smoothing function
        """    
        return (lambda t: np.where(t < 0 , 0, np.where( t <= 1 , 3*t**2-2*t**3, 1 ) ) )( (x-mi)/(mx-mi) )


class TimeStampedAudioSignal(AudioSignal):
    """ Audio signal with global time stamp and optionally a tag 
        that may be used to indicate the source.

        Args:
            rate: int
                Sampling rate in Hz
            data: 1d numpy array
                Audio data 
            time_stamp: datetime
                Global time stamp marking start of audio recording
            tag: str
                Optional argument that may be used to indicate the source.
    """

    def __init__(self, rate, data, time_stamp, tag=""):
        AudioSignal.__init__(self, rate, data, tag)
        self.time_stamp = time_stamp

    @classmethod
    def from_audio_signal(cls, audio_signal, time_stamp, tag=""):
        """ Initialize time stamped audio signal from regular audio signal.

            Args:
                audio_signal: AudioSignal
                    Audio signal
                time_stamp: datetime
                    Global time stamp marking start of audio recording
                tag: str
                    Optional argument that may be used to indicate the source.
        """
        return cls(audio_signal.rate, audio_signal.data, time_stamp, tag)

    @classmethod
    def from_wav(cls, path, time_stamp):
        """ Generate time stamped audio signal from wave file

            Args:
                path: str
                    Path to input wave file
                time_stamp: datetime
                    Global time stamp marking start of audio recording

            Returns:
                Instance of TimeStampedAudioSignal
                    Time stamped audio signal from wave file
        """        
        signal = super(TimeStampedAudioSignal, cls).from_wav(path=path)
        return cls.from_audio_signal(audio_signal=signal, time_stamp=time_stamp)

    def copy(self):
        """ Makes a copy of the time stamped audio signal.

            Returns:
                Instance of TimeStampedAudioSignal
                    Copied signal
        """                
        signal = super(TimeStampedAudioSignal, self).copy()
        return self.from_audio_signal(audio_signal=signal, time_stamp=self.time_stamp)

    def begin(self):
        """ Get global time stamp marking the start of the audio signal.

            Returns:
                t: datetime
                Global time stamp marking the start of audio the recording
        """
        t = self.time_stamp
        return t

    def end(self):
        """ Get global time stamp marking the end of the audio signal.

            Returns:
                t: datetime
                Global time stamp marking the end of audio the recording
        """
        duration = len(self.data) / self.rate
        delta = datetime.timedelta(seconds=duration)
        t = self.begin() + delta
        return t 

    def _crop(self, i1, i2):
        """ Crop time-stamped audio signal using [i1, i2] as cropping range

            Args:
                i1: int
                    Lower bound of cropping interval
                i2: int
                    Upper bound of cropping interval
        """   
        dt = max(0, i1/self.rate)
        dt = min(self.duration(), dt)
        self.time_stamp += datetime.timedelta(microseconds=1e6*dt) # update time stamp

        super(TimeStampedAudioSignal, self)._crop(i1, i2)   # crop signal
        
    def crop(self, begin=None, end=None):
        """ Crop time-stamped audio signal

            Args:
                begin: datetime
                    Start date and time of selection window
                end: datetime
                    End date and time of selection window
        """   
        b, e = None, None
        
        if begin is not None:
            b = (begin - self.begin()).total_seconds()
        if end is not None:
            e = (end - self.begin()).total_seconds()

        i1, i2 = self._selection(b, e)
        
        self._crop(i1, i2)

    def split(self, s):
        """ Split time-stamped audio signal.

            After splitting, this instance contains the remaining part of the audio signal.        

            Args:
                s: int
                    If s >= 0, select samples [:s]. If s < 0, select samples [-s:]
                    
            Returns:
                Instance of TimeStampedAudioSignal
                    Selected part of the audio signal.
        """   
        if s is None:
            s = len(self.data)

        if s >= 0:
            t = self.begin()
        else:
            dt = -s / self.rate
            t = self.end() - datetime.timedelta(microseconds=1e6*dt) # update time stamp
            
        a = super(TimeStampedAudioSignal, self).split(s)
        return self.from_audio_signal(audio_signal=a, time_stamp=t)

    def append(self, signal, delay=None, n_smooth=0, max_length=None):
        """ Combine two time-stamped audio signals.

            If delay is None (default), the delay will be determined from the 
            two audio signals' time stamps.

            See :meth:`audio_signal.append` for more details.

            Args:
                signal: AudioSignal
                    Audio signal to be merged
                delay: float
                    Delay between the two audio signals in seconds.
                    
            Returns:
                t: datetime
                    Start time of appended part.
        """   
        if delay is None:
            delay = self.delay(signal)

        dt = super(TimeStampedAudioSignal, self).append(signal=signal, delay=delay, n_smooth=n_smooth, max_length=max_length)
        t = self.begin() + datetime.timedelta(microseconds=1e6*dt)
        return t

    def delay(self, signal):
        """ Compute delay between two time stamped audio signals, defined 
            as the time difference between the end of the first signal and 
            the beginning of the second signal.

            Args:
                signal: TimeStampedAudioSignal
                    Audio signal

            Returns:
                d: float
                    Delay in seconds
        """   
        d = None
        
        if isinstance(signal, TimeStampedAudioSignal):
            d = (signal.begin() - self.end()).total_seconds()
        
        return d



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'cnn' module within the ketos library
"""

import pytest
import os
import tables
import numpy as np
import ketos.data_handling.data_handling as dh
import ketos.data_handling.database_interface as di
from ketos.data_handling.data_feeding import BatchGenerator, ActiveLearningBatchGenerator
from ketos.neural_networks.cnn import BasicCNN, ConvParams
from tensorflow import reset_default_graph

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')

@pytest.mark.test_BasicCNN
def test_initialize_BasicCNN_with_default_constructor_and_default_args(database_prepared_for_NN):
    d = database_prepared_for_NN
    train_x = d["train_x"]
    train_y = d["train_y"]
    _ = BasicCNN(train_x=train_x, train_y=train_y, verbosity=0)
    reset_default_graph()

@pytest.mark.test_BasicCNN
def test_train_BasicCNN_with_default_args(database_prepared_for_NN):
    d = database_prepared_for_NN
    train_x = d["train_x"]
    train_y = d["train_y"]
    validation_x = d["validation_x"]
    validation_y = d["validation_y"]
    test_x = d["test_x"]
    test_y = d["test_y"]
    network = BasicCNN(train_x=train_x, train_y=train_y, validation_x=validation_x, validation_y=validation_y, test_x=test_x, test_y=test_y, num_labels=2, verbosity=0)
    _ = network.create()
    network.train()
    reset_default_graph()

@pytest.mark.test_BasicCNN
def test_train_BasicCNN_with_two_channel_images():
    N = 32
    # training data
    train_x = np.random.randn(N,8,6,2) + 3 * np.random.uniform()
    train_y = np.sum(train_x, axis=(1,2,3)) > 1.5
    # initialize network
    network = BasicCNN(train_x=train_x, train_y=train_y,  num_labels=2, verbosity=0)
    _ = network.create()
    network.train()
    reset_default_graph()

@pytest.mark.test_BasicCNN
def test_train_BasicCNN_with_two_channel_images_and_validation():
    N = 128
    # training data
    train_x = np.random.randn(N,8,6,2) + 3 * np.random.uniform()
    train_y = np.sum(train_x, axis=(1,2,3)) > 1.5
    # validation data
    val_x = np.random.randn(N,8,6,2) + 3 * np.random.uniform()
    val_y = np.sum(val_x, axis=(1,2,3)) > 1.5
    # initialize network
    network = BasicCNN(train_x=train_x, train_y=train_y, validation_x=val_x, validation_y=val_y,\
            num_labels=2, verbosity=0, batch_size=32, num_epochs=10)
    _ = network.create()
    network.train()
    reset_default_graph()

@pytest.mark.test_BasicCNN
def test_train_BasicCNN_with_batch_norm():
    x = 2.0 * np.random.randn(128,8,8) + 1.5
    y = np.random.randn(128)
    y = (y > 0.5)    
    network = BasicCNN(train_x=x, train_y=y, num_labels=2, verbosity=0, batch_size=32, num_epochs=3)
    _ = network.create(batch_norm=True)
    network.train()
    # retrieve moving average and variance
    import tensorflow as tf
    with tf.variable_scope("", reuse=tf.AUTO_REUSE):
        out = network.sess.run([tf.get_variable('batch_normalization/moving_mean'),
                        tf.get_variable('batch_normalization/moving_variance')])
        moving_average, moving_variance = out
        assert moving_average[0] == pytest.approx(0.2, abs=0.1)
        assert moving_variance[0] == pytest.approx(1.3, abs=0.2)

    reset_default_graph()

@pytest.mark.test_BasicCNN
def test_train_BasicCNN_with_default_args2(database_prepared_for_NN_2_classes):
    d = database_prepared_for_NN_2_classes
    train_x = d["train_x"]
    train_y = d["train_y"]
    validation_x = d["validation_x"]
    validation_y = d["validation_y"]
    test_x = d["test_x"]
    test_y = d["test_y"]
    network = BasicCNN(train_x=train_x, train_y=train_y, validation_x=validation_x, validation_y=validation_y, test_x=test_x, test_y=test_y, num_labels=2, verbosity=0)
    _ = network.create()
    network.train()
    reset_default_graph()

@pytest.mark.test_BasicCNN
def test_train_BasicCNN_with_train_batch_generator(database_prepared_for_NN_2_classes):
    """ Test if batch generator returns labels instead of boxes
    """
    database = tables.open_file(os.path.join(path_to_assets,"humpback.h5"), 'r')
    train_data = di.open_table(database, "/train/mel_specs" )
    val_data = di.open_table(database, "/validation/mel_specs" )
    test_data = di.open_table(database, "/test/mel_specs")
    
    image_shape = train_data[0]['data'].shape
    def parse_y(y):
        labels=list(map(lambda l: dh.to1hot(di.parse_labels(l)[0], depth=2), y ))
        return np.array(labels)


    def apply_to_batch(X,Y):
        Y = parse_y(Y)
        return (X,Y)
    
    validation_x = val_data[:5]['data']
    validation_y = parse_y(val_data[:5]['labels'])
    
    train_generator = BatchGenerator(hdf5_table=train_data, y_field='labels', batch_size=5, return_batch_ids=False, instance_function=apply_to_batch) 
    
       
    network = BasicCNN(image_shape=image_shape, validation_x=validation_x, validation_y=validation_y, num_epochs=2, num_labels=2, seed=123, batch_size=5, verbosity=0 )

    conv_params = [ConvParams(name='conv_1', n_filters=32, filter_shape=[2,8]),
         ConvParams(name='conv_2', n_filters=64, filter_shape=[30,8])]

    _ = network.create(conv_params=conv_params, dense_size=[512])
    network.train(train_batch_gen=train_generator)
    # network.train()
    reset_default_graph()

@pytest.mark.test_BasicCNN
def test_load_BasicCNN_model(database_prepared_for_NN_2_classes, trained_BasicCNN):
    d = database_prepared_for_NN_2_classes
    train_x = d["train_x"]
    train_y = d["train_y"]
    validation_x = d["validation_x"]
    validation_y = d["validation_y"]
    test_x = d["test_x"]
    test_y = d["test_y"]
    network = BasicCNN(train_x=train_x, train_y=train_y, validation_x=validation_x, validation_y=validation_y, test_x=test_x, test_y=test_y, num_labels=2, verbosity=0)
    path_to_meta, path_to_saved_model, test_acc = trained_BasicCNN
    _ = network.load(path_to_meta, path_to_saved_model)
    assert test_acc == network.accuracy_on_test()
    reset_default_graph()
    
@pytest.mark.test_BasicCNN
def test_compute_class_weights_with_BasicCNN(database_prepared_for_NN_2_classes):
    d = database_prepared_for_NN_2_classes
    x = d["train_x"]
    y = d["train_y"]
    network = BasicCNN(train_x=x, train_y=y, num_labels=2, verbosity=0, seed=41)
    _ = network.create()
    network.train()
    img = np.zeros((20, 20))
    result = network.get_class_weights(x=[img])
    weights = result[0]
    assert weights[0] + weights[1] == pytest.approx(1.000, abs=0.001)
    reset_default_graph()

@pytest.mark.test_BasicCNN
def test_compute_class_weights_with_BasicCNN_with_batch_norm(database_prepared_for_NN_2_classes):
    d = database_prepared_for_NN_2_classes
    x = d["train_x"]
    y = d["train_y"]
    network = BasicCNN(train_x=x, train_y=y, num_labels=2, verbosity=0, seed=41)
    _ = network.create(batch_norm=True)
    network.train()
    img = np.zeros((20, 20))
    result = network.get_class_weights(x=[img])
    weights = result[0]
    assert weights[0] + weights[1] == pytest.approx(1.000, abs=0.001)
    reset_default_graph()

@pytest.mark.test_BasicCNN
def test_compute_features_with_BasicCNN(database_prepared_for_NN_2_classes):
    d = database_prepared_for_NN_2_classes
    x = d["train_x"]
    y = d["train_y"]
    network = BasicCNN(train_x=x, train_y=y,  num_labels=2, verbosity=0, seed=41)
    _ = network.create()
    network.train()
    img = np.zeros((20, 20))
    result = network.get_features(x=[img], layer_name='dense_1')
    f = result[0]
    assert f.shape == (512,)
    reset_default_graph()

@pytest.mark.test_BasicCNN
def test_active_learning():
    # initialize BasicCNN for classifying 2x2 images
    cnn = BasicCNN(image_shape=(2,2), verbosity=0, seed=1)
    # create a small network with one convolutional layers and one dense layer
    params = ConvParams(name='conv_1', n_filters=4, filter_shape=[2,2])
    _ = cnn.create(conv_params=[params], dense_size=[4])
    # create some training data
    img0 = np.zeros(shape=(2,2))
    img1 = np.ones(shape=(2,2))
    x = [img0, img1, img0, img1, img0, img1, img0] # input data
    y = [0, 1, 0, 1, 0, 1, 0] # labels
    # create a data provider
    g = ActiveLearningBatchGenerator(session_size=4, batch_size=2, x=x, y=y)
    # train it
    _, _ = cnn.train_active(provider=g, num_sessions=3, num_epochs=7, learning_rate=0.005)
    reset_default_graph()